--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]
--
--[[
-- This module defines the combatstate. The combatstate is where we track
-- all the state about combat.
--]]

local M = {}
local log = require 'log'
local scarlettables = require 'scarlettables'

local function secondarystatistics(stats)
    local secondarystatistics = {
        attackmodifiers = {},
        defensemodifiers = {},
        shield={damage=0},
        retaliation={damage=0}
    }
    for _, statistic in ipairs(stats) do
        secondarystatistics[statistic .. "modifiers"] = {}
    end
    return secondarystatistics
end

-- [[
-- Builds an initial combat state. See `combat.md` for a description of the
-- canonical mappings of a combatstate.
--
-- gamestate - The current state of the game when combat begins
-- victory - The page to transition to if the player is victorious
-- loss - The page to transition to if the player is defeated. If `nil`,
--  then the player must win to advance the game
-- allies - The list of allies in combat
-- enemies - The list of enemies in combat
--
--  returns - A combatstate for the start of combat
-- ]]
function M.combatstate(gamestate, victory, loss, allies, enemies)
    local combatallies = scarlettables.copy(allies)
    local combatenemies = scarlettables.copy(enemies)
    local characters = M.concatenate(combatallies, combatenemies)
    for _, character in pairs(characters) do
    end
    local combatstate = {
        start={allies=allies, enemies=enemies},
        round=1,
        victory=victory,
        defeat=loss,
        allies=combatallies,
        enemies=combatenemies,
        characters=characters,
        charindex=1,
        log={},
        selections={},
        defeated={},
        choicecounts={},
        combatlog={},
        -- A bag for statuses to use if they need to remember something other
        -- than stat modifiers (like original action counts).
        statusstate={},
        metalog={},
        attrition=gamestate.attrition,
        ai=gamestate.options.ai
    }
    for _, character in pairs(characters) do
        character.secondarystatistics = secondarystatistics(
            character.statistics._keys
        )
        character.attrition=gamestate.attrition(character)
        character.statuses = {} 
        local choicecounts = {_keys={}}
        for _, choicename in ipairs(character.choicecounts._keys) do
            local choice = character.choicecounts[choicename]
            choicecounts[choicename] = {
                current=choice.count,
                max=choice.count,
                page=choice.page
            }
            table.insert(choicecounts._keys, choicename)
        end
        combatstate.choicecounts[character.name] = choicecounts
    end
    combatstate.choice = require 'combat/combat'.firstpage(combatstate)
    return combatstate
end

-- [[
-- Checks if the passed in character is an ally.
--
-- combatstate: The current state of combat
-- character: The character who may be an ally
--
-- return True if the character is an ally, False if the character is an enemy,
-- fails if the character is neither
-- ]]
function M.isally(combatstate, character)
    for _, ally in pairs(combatstate.allies) do
        if ally == character then
            return true
        end
    end
    for _, enemy in pairs(combatstate.enemies) do
        if enemy == character then
            return false
        end
    end
    assert(
        false,
        "Character: " .. scarlettables.string(character) .. " is not in combat."
    )
end

--[[
-- Checks if the first character is an ally of the second.
--
-- combatstate: The current state of combat
-- character1: One of the characters who may be allies
-- character2: The other character who may be allies
--
-- return True if the two characters are allies, false otherwise
--]]
function M.areallies(combatstate, character1, character2)
    local character1allies = M.isally(combatstate, character1)
        and combatstate.allies
        or combatstate.enemies
    for _, character in ipairs(character1allies) do
        if character == character2 then
            return true
        end
    end
    return false
end

-- [[
-- Returns the currently active character
--
-- combatstate: The current state of combat
--
-- return the character whose actions are currently being selected
-- ]]
function M.activecharacter(combatstate)
    local charindex = combatstate.charindex
    local character = combatstate.characters[charindex]
    return assert(character, "No character at " .. charindex .. ".")
end

function M.concatenate(allies, enemies)
    local characters = {}
    for i, character in pairs(allies) do
        characters[i] = character
    end
    local numallies = #characters
    for i, character in pairs(enemies) do
        characters[i + numallies] = character
    end
    return characters
end

--[[
-- Given the name of a character, returns the copy of that character stored
-- in this combatstate.
--
-- charactername - The name of the character to find
--
-- return The character with that name, and the index of that character in the
-- list of all characters in combat
--]]
function M.findcharacter(combatstate, charactername)
    for i, character in pairs(combatstate.characters) do
        if character.name == charactername then
            return character, i
        end
    end
end

--[[
-- Given the combatname of a character, returns the first character
-- in combat with that name.
--
-- combatstate: The current state of combat
-- combatname: The combat name of the character desired
--
-- return The first character found with this combatname, or nil
-- if there is no character with this combatname
--]]
function M.findcharacterbycombatname(combatstate, combatname)
    for _, character in ipairs(combatstate.characters) do
        if character.combatname == combatname then
            return character
        end
    end
end

--[[
-- Marks the specified character as defeated.
--
-- If the character is a summon, then we also remove them from
-- the character list. This allows us to reuse the defeated character's
-- index when summoning more.
--
-- charactername - The name of the character to defeat
--]]
function M.defeatcharacter(combatstate, charactername)
    if M.isdefeated(combatstate, charactername) then
        return
    end
    local character, index = M.findcharacter(combatstate, charactername)
    combatstate.defeated[character.name] = character
end

--[[
-- Revives a character, essentially unmarking them as defeated.
--]]
function M.revivecharacter(combatstate, charactername)
    combatstate.defeated[charactername] = nil
    local choices = combatstate.choicecounts[charactername]
    for _, choicename in ipairs(choices._keys) do
        local choice = choices[choicename]
        choice.current = choice.max
    end
end

--[[
-- If a character with the specified name has been defeated, returns that
-- character. Otherwise, returns `nil`
--
-- combatstate - The current state of combat
-- charactername - The name of the character who may or may not be defeated
--
-- return - The character if they have been defeated, `nil` otherwise
-- ]]
function M.isdefeated(combatstate, charactername)
    return combatstate.defeated[charactername]
end

--[[
-- Given a list of characters, returns true if all of those characters have
-- been defeated, and false otherwise
--
-- combatstate: The current state of combat
-- characters: The characters to check
--
-- return True if all the characters have been defeated, false otherwise
--]]
function M.alldefeated(combatstate, characters)
    for _, character in pairs(characters) do
        if not M.isdefeated(combatstate, character.name) then
            return false
        end
    end
    return true
end

-- [[
-- Given a combatstate, and a character, returns the list of undefeated
-- characters who are that character's opponents.
--
-- combatstate: The current state of combat
-- character: A character whose opponents are desired
--
-- return The list of characters who are the specified character's opponents
-- ]]
function M.opponents(combatstate, character)
    local opponents
    for _, enemy in ipairs(combatstate.enemies) do
        if character == enemy then
            opponents = combatstate.allies
            break
        end
    end
    if not opponents then
        for _, ally in ipairs(combatstate.allies) do
            if character == ally then
                opponents = combatstate.enemies
                break
            end
        end
    end
    if not opponents then
        local errormessage = "This character is neither an ally or an enemy: "
            .. scarlettables.string(character)
        log.error(errormessage)
        assert(false, errormessage)
    end
    local activeopponents = {}
    local count = 1
    for _, opponent in ipairs(opponents) do
        if not combatstate.defeated[opponent.name] then
            activeopponents[count] = opponent
            count = count + 1
        end
    end
    return activeopponents
end

--[[
-- Returns the undefeated opponents of the active character
--
-- combatstate: The current state of combat
--
-- return The list of characters who are the active character's opponents
--]]
function M.activeopponents(combatstate)
    return M.opponents(combatstate, M.activecharacter(combatstate))
end


-- [[
-- Given a combatstate, and a character, returns the list of undefeated
-- characters who are that character's allies.
--
-- combatstate: The current state of combat
-- character: A character whose allies are desired
--
-- return The list of characters who are the specified character's allies
-- ]]
function M.allies(combatstate, character)
    local allies
    for _, enemy in ipairs(combatstate.enemies) do
        if character == enemy then
            allies = combatstate.enemies
            break
        end
    end
    if not allies then
        for _, ally in ipairs(combatstate.allies) do
            if character == ally then
                allies = combatstate.allies
                break
            end
        end
    end
    if not allies then
        local errormessage = "This character is neither an ally or an enemy: "
            .. scarlettables.string(character)
            .. "\nCombatstate: " .. scarlettables.string(combatstate)
        log.error(errormessage)
        assert(false, errormessage)
    end
    local activeallies = {}
    for _, ally in ipairs(allies) do
        if not M.isdefeated(combatstate, ally.name) then
            table.insert(activeallies, ally)
        end
    end
    return activeallies
end

--[[
-- Returns whether or not the specified character is 
-- active (i.e. not defeated)
--
-- combatstate: The current state of combat
-- character: A table with at least
--  one key: name, mapping to the character's name
--
-- returns true if the character is active, false otherwise
--]]
function M.isactive(combatstate, character)
    return not combatstate.defeated[character.name]
end

-- [[
-- Returns all the currently active characters
-- ]]
function M.activecharacters(combatstate)
    local activecharacters = {}
    local count = 1
    for _, character in ipairs(combatstate.characters) do
        if not combatstate.defeated[character.name] then
            activecharacters[count] = character
            count = count + 1
        end
    end
    return activecharacters
end

--[[
-- Returns the undefeated allies of the active character
--
-- combatstate: The current state of combat
--
-- return The list of characters who are the active character's allies
--]]
function M.activeallies(combatstate)
    return M.allies(combatstate, M.activecharacter(combatstate))
end

local function addcharacter(combatstate, characters, newcharacter)
    if string.find(newcharacter.name, "_summon") then
        for i, character in ipairs(characters) do
            local name = character.name
            if M.isdefeated(combatstate, name) then
                characters[i] = newcharacter
                table.insert(characters, character)
                return
            end
        end
    end
    table.insert(characters, newcharacter)
end

function M.addally(combatstate, ally)
    if not ally.secondarystatistics then
        ally.secondarystatistics = secondarystatistics(ally.statistics._keys)
    end
    if not ally.statuses then
        ally.statuses = {}
    end
    if not ally.attrition then
        ally.attrition=combatstate.attrition(ally)
    end
    addcharacter(combatstate, combatstate.allies, ally)
    for index, character in ipairs(combatstate.characters) do
        if not M.isally(combatstate, character) then
            combatstate.characters[index] = ally
            table.insert(combatstate.characters, index+1, character)
            break
        end
    end
    local choicecounts = {_keys={}}
    for _, choicename in ipairs(ally.choicecounts._keys) do
        local choice = ally.choicecounts[choicename]
        choicecounts[choicename] = {
            current=choice.count,
            max=choice.count,
            page=choice.page
        }
        table.insert(choicecounts._keys, choicename)
    end
    combatstate.choicecounts[ally.name] = choicecounts
end

function M.addenemy(combatstate, enemy)
    if not enemy.secondarystatistics then
        enemy.secondarystatistics = secondarystatistics(enemy.statistics._keys)
    end
    if not enemy.statuses then
        enemy.statuses = {}
    end
    if not enemy.attrition then
        enemy.attrition=combatstate.attrition(enemy)
    end
    addcharacter(combatstate, combatstate.characters, enemy)
    addcharacter(combatstate, combatstate.enemies, enemy)
    local choicecounts = {_keys={}}
    for _, choicename in ipairs(enemy.choicecounts._keys) do
        local choice = enemy.choicecounts[choicename]
        choicecounts[choicename] = {
            current=choice.count,
            max=choice.count,
            page=choice.page
        }
        table.insert(choicecounts._keys, choicename)
    end
    combatstate.choicecounts[enemy.name] = choicecounts
    local charnames = {}
    for i, character in ipairs(combatstate.characters) do
        charnames[i] = character.name
    end
    log.info("Characters after adding enemy: " .. scarlettables.string(charnames))
    local enemynames = {}
    for i, enemy in ipairs(combatstate.enemies) do
        enemynames[i] = enemy.name
    end
    log.info("Enemies after adding enemy: " .. scarlettables.string(enemynames))
end

return M
