--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]
--
--[[
-- Defines the story page for combat. This module is responsible for guiding
-- the player through combat. As such, it contains the combatlogic for
-- transitioning through the combat pages.
--
-- However, in order to allow game developers to define combat pages in
-- isolation (i.e. without having to worry about which other choices they
-- should be chained to), the combat module uses the following algorithm
-- for transitioning through combat pages
-- (defined in combat/operations.nextcombatpage):
--
-- 1. If the combat page has choices or an automatic page, then it transitions
-- to one of those depending on the choices made by the player, or the state
-- of combat respectively.
--
-- 2. If the combat page has neither choices nor automatics, then the game
-- queries the combatstate for the next ally, and uses that ally's choice.
--
-- 3. If we have iterated through all of the allies, we next iterate through
-- the choices for all of the enemies.
--
-- 4. If we have iterated through all the allies and enemies, then the round
-- is over, and we get the next page from `combatdata.endround`, a function
-- that takes the current state of combat and returns a page that handles any
-- end of round actions and transitions to the next round.
--
-- If no `combatdata.endround` is defined, then we increment the round counter,
-- clear the combat action choices, drop the previous combat state (so you can't
-- back up past this round) and transition to the first ally's choice.
--
-- Note: We do NONE of these things automatically if a `endround` is provided.
--
-- If there are no longer any active allies or enemies then combat is ended.
--
-- When building choices, the key that corresponds to the choice must be wrapped
-- in parenthesis. For example in the choice string (A)ttack, the game
-- will interpret `a` as the key to press to attack. The only exception to this
-- is the pound (or number) sign `#`. This is used to represent any digit.
--
-- The engine will automatically expand that out into a choice for every
-- digit: 0 - 9, and each function key: F1 - F12. Each page delegates to the
-- appropriate function in the
-- page defined by the user, except that the transformer sets the mapping
-- `combatstate.target` to the pressed key first.
--
-- So, targeting actions can access the targeted character through the
-- `combatstate.target` function.
--]]

local M = {}
local combatcommands = require 'combat/commands'
local combatstateops = require 'combat/combatstate'
local combatoperations = require 'combat/operations'
local fonts = LOVE and require 'ui/fonts'
local log = require 'log'
local love = LOVE and require 'love'
local pages = require 'pages'
local saveload = require 'utilities/saveloadui'
local scarlettables = require 'scarlettables'
local uiutilities = require 'ui/utilities'

--[[
-- The player may go up against multiples of the same enemy. Therefore, we need
-- to make them unique. We do this by:
--
-- 1. Making a deep copy of each enemy.
-- 2. Appending a count to their name. So the first instance of each enemy
-- doesn't have a count, and each subsequent one has a count.
--
-- enemies: The enemies to make unique
--
-- return A deep copy of the specified enemies with their names made unique
--]]
local function uniquifyenemies(enemies)
    local uniqueenemies = {}
    local enemynames = {}
    for i, character in ipairs(enemies) do
        local copy = scarlettables.copy(character)
        -- Verify that our deep copy is in fact a deep copy.
        assert(copy.statistics ~= character.statistics)
        local numenemytypes = enemynames[character.name]
        if numenemytypes then
            copy.name = copy.name .. tostring(numenemytypes)
            enemynames[character.name] = numenemytypes + 1
        else
            enemynames[character.name] = 2
        end
        uniqueenemies[i] = copy
    end
    return uniqueenemies
end

--[[
-- Builds the first UI page for combat. This page displays the first choice
-- for combat (probably a choice for the player's first character), and
-- creates a mapping `combatstate` in `gamestate` that contains the current
-- state of combat.
--
-- gamestate:  The current state of the game
-- victory: The game page to use if the player wins
-- loss: The game page to use if the player loses
-- allies: A list of allies involved in the battle
-- enemies: A list of enemies involved in the battle
--
-- returns A UI page that displays the initial combat screen, and assigns
-- the `gamestate.combatstate` mapping to the initial combatstate.
-- ]]
function M.startcombat(gamestate, victory, loss, allies, enemies)
    local combatpage = combatoperations.startcombatpage(
        gamestate,
        victory,
        loss,
        allies,
        uniquifyenemies(enemies)
    )
    return M.combatuipage(
        combatpage,
        function(uistate)
            uiutilities.clearundo(uistate)
            uistate.line = 1
        end
    )
end

local TEXT_SCREEN_PERCENTAGE = 2 / 3

--[[
-- Given a paragraph, returns true if the paragraph has just empty lines,
-- false otherwise.
--
-- paragraph - The paragraph to check
--
-- returns true if the paragraph is blank, false otherwise
--]]
local function isblank(paragraph)
    for _, line in pairs(paragraph) do
        if string.match(line, "%S") then
            return false
        end
    end
    return true
end

-- [[
-- Given a collection of paragraphs, strips off the first and last if they
-- only contain blank lines.
--
-- paragraphs: The list of paragraphs to trim
-- ]]
local function trim(paragraphs)
    if isblank(paragraphs[1]) then
        table.remove(paragraphs, 1)
    end
    if isblank(paragraphs[#paragraphs]) then
        table.remove(paragraphs)
    end
end

--[[
-- Builds a label that displays the combat board
--
-- uistate - The current state of the UI. The UI must include the following
--  keys:
--      line - The current line in the text that we are beginning our display
--
--  love - The love game engine
--  font - The font to use to display the text
--  text - The text to display
--  screenratio - The amount of screen space the body should take up
-- ]]
local function buildboard(uistate, text, screenratio, maxlength)
    maxlength = maxlength or 3 * M.MAX_LINE_LENGTH
    local options = uistate.core.gamestate.options
    local charwidth = fonts.singleCharacterWidth(fonts.body(options.bodyfont))
    local screenwidth, screenheight = love.graphics.getDimensions()
    local numcharsline = (screenwidth - charwidth) / charwidth
    local labelheight = screenheight * screenratio
    local storyutils = require 'story/storyutils'
    local paragraphs = storyutils.wraptext(text, numcharsline, maxlength)

    trim(paragraphs)
    for i, paragraph in ipairs(paragraphs) do
        paragraphs[i] = table.concat(paragraph, "\n")
    end

    local pixellength = numcharsline * charwidth
    local x, y, w, h = uistate.gui.layout:row(pixellength, labelheight)
    local padding_x, padding_y = uistate.gui.layout:padding()
    uistate.gui.layout:padding(padding_x / 2, padding_y / 2)
    uistate.gui.Label(
        table.concat(paragraphs, "\n"),
        {
            id="story",
            font=fonts.combatboard(options.combatfont),
            align="left",
            valign="top"
        },
        x, y, w, h
    )
    uistate.gui.layout:padding(padding_x, padding_y)
end

--[[
-- Draws the board showing all the characters and their statuses in combat.
--
-- uistate: The current state of the UI
-- maxlinelength: The maximum number of characters allowed on a single line
--]]
local function drawboard(uistate, maxlinelength)
    local combatstate = uistate.core.gamestate.combatstate
    local enemies = M.drawcharacters(combatstate, combatstate.enemies)
    enemies = table.concat(enemies, " ")
    local allies = M.drawcharacters(combatstate, combatstate.allies, "F")
    allies = table.concat(allies, " ")

    local activecharacter = combatstate.characters[combatstate.charindex]
    local activecharactername = assert(
        activecharacter.combatname,
        "No combat name for character: " .. activecharacter.name
    )
    local board = {
        " ---- Enemies ---- ",
        enemies,
        " ---- Allies ---- ",
        allies,
        "--- Round " .. combatstate.round .. " ---"
    }
    buildboard(uistate, board, 1, maxlinelength)
end

--[[
-- Given a UI state, and the name of the desired page, builds the choices to
-- display to the user
--
-- uistate: The current state of the ui
-- allowsaving: Whether or not to allow saving, defaults to true
--]]
local function buildcombatchoices(uistate, uipage)
    local gui = uistate.gui
    local buttonfont = fonts.button(uistate.core.gamestate.options.buttonfont)
    local backtext = "(<==)Back"
    local commands = "(Spacebar)Commands"
    local buttonwidth = buttonfont:getWidth(commands) + 5
    gui.Button(
        commands,
        {id="space", font=buttonfont},
        gui.layout:row(buttonwidth, buttonfont:getHeight())
    )
    gui.Button(
        "(.)Actions",
        {id=".", font=buttonfont},
        gui.layout:col()
    )
    gui.Button(
        backtext,
        {id="backspace", font=buttonfont},
        gui.layout:col()
    )
end

local function consolecommands(uistate, uipage)
    local commands = {}
    if uistate.core.gamestate.combatstate.previous then
        table.insert(commands, "(<) Back")
    end
    table.insert(commands, "(e)nd Turn")
    table.insert(commands, "(r)estart")
    table.insert(commands, "(l)oad")
    table.insert(commands, "(t)itle")
    local lines = {""}
    local numcommands = #commands
    for i = 1,numcommands,3 do
        local line = {
            commands[i],
            i+1 <= numcommands and commands[i+1] or "",
            i+2 <= numcommands and commands[i+2] or ""
        }
        table.insert(lines, table.concat(line, "\t"))
    end
    print(table.concat(lines, "\n"))
end

local function consoleboard(uistate)
    local combatstate = uistate.core.gamestate.combatstate
    local combatstate = uistate.core.gamestate.combatstate
    local activecharacters = combatstateops.activecharacters(combatstate)
    local charstrings = M.drawcharacters(combatstate, combatstate.characters)
    local enemies = {}
    local allies = {} 
    for i, character in ipairs(combatstate.characters) do
        local characterstring = charstrings[i]
        if characterstring and string.len(characterstring) > 0 then
            if combatstateops.isally(combatstate, character) then
                table.insert(allies, characterstring)
            else
                table.insert(enemies, characterstring)
            end
        end
    end 
    enemies = table.concat(enemies, "\n")
    allies = table.concat(allies, "\n")
    local board = {
        " ---- Allies ---- ",
        allies,
        " ---- Enemies ---- ",
        enemies,
        ""
    }
    print(table.concat(board, "\n"))
end

local function consoleactions(uistate)
    local combatstate = uistate.core.gamestate.combatstate
    local activecharacter = combatstate.characters[combatstate.charindex]
    local activecharactername = assert(
        activecharacter.combatname,
        "No combat name for character: " .. activecharacter.name
    )
    local actions = {"---- Active: " ..  activecharactername .. " ----"}
    for _, line in ipairs(combatstate.metalog) do
        table.insert(actions, line)
    end
    print(table.concat(actions, "\n"))
end

--[[
-- Prints a UI for playing on the console to stdout.
--
-- uistate: The current state of the game
-- body: The list of strings to print
-- uipage: The current page for combat
--]]
local function consoleui(uistate, body, uipage)
    local storyutils = require 'story/storyutils'

    local wrappedtext = storyutils.wraptext(body, 80)
    local stars = string.rep("*", 80)
    local paragraphs = {stars}
    for _, lines in ipairs(wrappedtext) do
        table.insert(paragraphs, table.concat(lines, "\n"))
    end


    local combatstate = uistate.core.gamestate.combatstate
    local manydashes = string.rep("-", 35)
    print(table.concat(paragraphs, "\n"))
    consoleactions(uistate) 
    print(stars)

    consoleboard(uistate)
    consolecommands(uistate, uipage)
end

local function bottompage(uistate, maximumlines)
    uistate.scrollToBottom = false
    uistate.line = math.max(1, maximumlines - uistate.maxlinesonscreen)
end

--[[
-- Draws the UI for combat
--
-- uistate - The current state of the UI
-- uipage - The page to draw
--]]
local function drawgui(uistate, uipage)
    local storyutils = require 'story/storyutils'
    local body = {}
    for _, line in ipairs(uipage.body(uistate)) do
        if #line > 0 then
            table.insert(body, line)
        end
    end
    if LOVE then
        local combatstate = uistate.core.gamestate.combatstate
        local activecharacter = combatstate.characters[combatstate.charindex]
        local activecharactername = assert(
            activecharacter.combatname,
            "No combat name for character: " .. activecharacter.name
        )
        table.insert(body, "------- " .. activecharactername .. "------")
        for _, line in ipairs(combatstate.metalog) do
            table.insert(body, line)
        end
        local bodyfont = fonts.body(uistate.core.gamestate.options.bodyfont)
        local charwidth = fonts.singleCharacterWidth(bodyfont)
        local screenwidth = love.graphics.getDimensions()
        local numcharsline = math.min(
            (screenwidth - charwidth) / (2 * charwidth),
            storyutils.MAX_LINE_LENGTH
        )
        uistate.gui.layout:push(100, 100)
        storyutils.buildBodyLabel(
            uistate,
            love,
            fonts,
            body,
            {},
            TEXT_SCREEN_PERCENTAGE,
            numcharsline
        )
        if uistate.scrollToBottom then
            bottompage(uistate, uistate.numlines)
        end
        buildcombatchoices(uistate)
        uistate.gui.layout:pop()
        uistate.gui.layout:push(numcharsline * 2 * charwidth / 2, 100)
        drawboard(uistate, numcharsline)
        uistate.gui.layout:pop()
    else
        consoleui(uistate, body, uipage)
    end
end

local function decrementline(uistate)
    uistate.scrollToBottom = false
    uistate.line = math.max(1, uistate.line - 1)
end

local function decrementpage(uistate)
    uistate.scrollToBottom = false
    if uistate.textbefore then
        uistate.line = math.max(1, uistate.line - uistate.maxlinesonscreen)
    end
end

local function toppage(uistate)
    uistate.scrollToBottom = false
    uistate.line = 1
end

local function incrementline(uistate, maximumlines)
    uistate.scrollToBottom = false
    uistate.line = math.min(maximumlines, uistate.line + 1)
end

local function incrementpage(uistate, maximumlines)
    uistate.scrollToBottom = false
    local topbottom = math.max(1, maximumlines - uistate.maxlinesonscreen)
    if uistate.textafter then
        uistate.line = math.min(
            topbottom,
            uistate.line + uistate.maxlinesonscreen
        )
    end
end

--[[
-- Given a string representing a choice to make, returns the key to use for
-- that choice. If the choice contains a single letter wrapped in parenthesis,
-- then this function returns the Love KeyConstant associated with that letter,
-- or nil if there is no such letter.
--
-- choice: The choice to extract a key for
--
-- return The key wrapped in parenthesis in this choice, or nil if there is
-- no such key
--]]
local function extractkey(choice)
    local startparens = string.find(choice, "%(")
    local endparens = string.find(choice, "%)")
    if startparens and endparens then
        return string.sub(choice, startparens + 1, endparens - 1)
    end
end

local function characternames(characters)
    local names = {}
    for i, character in ipairs(characters) do
        names[i] = character.name
    end
    return names
end

--[[
-- Builds the map of button ids to pages that represent this page's
-- choices.
--
-- uistate - The current state of the UI
-- gamepage - The gamepage to build choices for
-- combatuipage - The combat UI page that we're building choices for
--]]
local function choiceids(uistate, gamepage, combatuipage)
    local uichoices = {_keys={}}
    local gamechoices = gamepage.choices(uistate.core.gamestate)
    -- This is passed to the combatcommands.actions function so
    -- that it knows which choices to expose as buttons.
    local actions = {_keys={}}
    local combatstate = uistate.core.gamestate.combatstate
    for _, choice in ipairs(gamechoices._keys) do
        -- For combat, if a command isn't wrapped in parenthesis, then
        -- it is a plain number or function key that should be used as the key.
        local key = string.lower(extractkey(choice) or choice)
        uichoices[key] = M.combatuipage(gamechoices[choice])
        table.insert(uichoices._keys, key)
        if not combatoperations.alphabetset[key] then
            if key == "return" then
                actions[key] = choice
                table.insert(actions._keys, key)
            else
                local character = string.match(key, "f%d%d?") and
                    combatstate.allies[tonumber(string.sub(key, 2))]
                    or combatstate.enemies[tonumber(key)]
                if character then
                    actions[key] = "(" .. key .. ") " .. character.combatname
                    table.insert(actions._keys, key)
                end
            end
        else
            local character = assert(
                combatstate.characters[combatstate.charindex],
                "No character at index: " .. combatstate.charindex ..
                " Characters: " .. scarlettables.string(characternames(combatstate.characters))
            )
            -- Happens if we're using alphanumeric characters to select powers.
            if key == choice and character.powers then
                local index = combatoperations.alphabetset[key]
                local powername = character.powers._keys[index]
                if powername then
                    actions[key] = "(" .. key .. ") " .. powername
                    table.insert(actions._keys, key)
                end
            else
                actions[key] = choice
                table.insert(actions._keys, key)
            end
        end
    end

    local function nextpage()
            uistate.scrollToBottom = false
            local combatpage = combatuipage
            combatpage.transformer = function() end
            return combatpage
    end
    if LOVE then
        local up = "up"
        uichoices[up] = pages.page(nil, nextpage, nil, decrementline)
        table.insert(uichoices._keys, up)

        local pageup = "pageup"
        uichoices[pageup] = pages.page(nil, nextpage, nil, decrementpage)
        table.insert(uichoices._keys, pageup)

        local home = "home"
        uichoices[home] = pages.page(nil, nextpage, nil, toppage)
        table.insert(uichoices._keys, home)

        local down = "down"
        uichoices[down] = pages.page(
            nil,
            nextpage,
            nil,
            function(uistate1)
                incrementline(uistate1, uistate1.numlines)
            end
        )
        table.insert(uichoices._keys, down)

        local pagedown = "pagedown"
        uichoices[pagedown] = pages.page(
            nil,
            nextpage,
            nil,
            function(uistate1)
                incrementpage(uistate1, uistate1.numlines)
            end
        )
        table.insert(uichoices._keys, pagedown)

        local endkey = "end"
        uichoices[endkey] = pages.page(
            nil,
            nextpage,
            nil,
            function(uistate1)
                bottompage(uistate1, uistate1.numlines)
            end
        )
        table.insert(uichoices._keys, endkey)
    end

    local r = "r"
    local gamestate = uistate.core.gamestate
    uichoices[r] = M.startcombat(
        gamestate,
        gamestate.combatstate.victory,
        gamestate.combatstate.defeat,
        gamestate.combatstate.start.allies,
        gamestate.combatstate.start.enemies
    )
    table.insert(uichoices._keys, r)

    local l = "l"
    uichoices[l] = saveload.saveload()
    table.insert(uichoices._keys, l)

    local titlescreen = require 'utilities/titlescreen'
    local t = "t"
    uichoices[t] = titlescreen.confirm(
        function()
            local combatpage = M.combatuipage(gamepage)
            combatpage.transformer = function() end
            return combatpage
        end
    )
    table.insert(uichoices._keys, t)

    local e = "e"
    uichoices[e] = M.combatuipage(
        combatoperations.page(
            combatoperations.nextcharacterpage(
                uistate.core.gamestate.combatstate
            )
        )
    )
    table.insert(uichoices._keys, e)

    if LOVE then
        local tab = "tab"
        uichoices[tab] = pages.page(
            nil,
            nextpage,
            nil,
            function(uistate1)
                local combatstate = uistate1.core.gamestate.combatstate
                combatstate.concise = not combatstate.concise
            end
        )
        table.insert(uichoices._keys, tab)

        local equals = "="
        uichoices[equals] = pages.page(
            nil,
            nextpage,
            nil,
            function(uistate1)
                local shiftheld = love.keyboard.isDown("lshift")
                    or love.keyboard.isDown("rshift")
                local options = uistate1.core.gamestate.options
                if shiftheld and options.combatfont then
                    options.combatfont = options.combatfont + 1
                end
            end
        )
        table.insert(uichoices._keys, equals)

        local minus = "-"
        uichoices[minus] = pages.page(
            nil,
            nextpage,
            nil,
            function(uistate1)
                local options = uistate1.core.gamestate.options
                if options.combatfont then
                    options.combatfont = options.combatfont - 1
                end
            end
        )
        table.insert(uichoices._keys, minus)

        uichoices.space = combatcommands.commands(uichoices, combatuipage)
        table.insert(uichoices._keys, "space")

        uichoices["."] = combatcommands.actions(uichoices, actions, combatuipage)
        table.insert(uichoices._keys, ".")
    end
    
    return uichoices
end

--[[
-- Builds a UI page from the passed in combat gamepage.
--
-- gamepage - The gamepage handling combat
-- transformer - An additional transformer to apply. This is applied *after*
-- the combat page's transformer. May be nil.
-- backup - Whether or not to create a previous copy of the combatstate for
-- backing up, defaults to true
--
-- return A UI page that executes the passed in gamepage
--]]
function M.combatuipage(gamepage, transformer, backup)
    if backup == nil then
        backup = true
    end
    -- Only story pages have names
    if gamepage.name then
        return require 'story/storyui'.page(gamepage.name, transformer)
    end
    local uipage = pages.page()
    function uipage.body(uistate)
        return gamepage.body(uistate.core.gamestate)
    end
    function uipage.automatics(uistate)
        local nextgamepage = gamepage.automatics(uistate.core.gamestate)
        return nextgamepage and M.combatuipage(nextgamepage, nil, false)
    end
    function uipage.choices(uistate)
        local choices = choiceids(uistate, gamepage, uipage)
        if uistate.core.gamestate.combatstate.previous then
            local backpage = pages.page(
                nil,
                function(gamestate)
                    local combatpage = gamestate.combatstate.currentpage
                    return combatoperations.page(combatpage)
                end,
                nil,
                function(gamestate)
                    local combatstate = gamestate.combatstate
                    if combatstate.previous then
                        log.info("Restoring backup.")
                        gamestate.combatstate = combatstate.previous
                    end
                end
            )
            local backkey = LOVE and "backspace" or "<"
            choices[backkey] = M.combatuipage(backpage, nil, false)
            table.insert(choices._keys, "backspace")
        end
        return choices
    end
    function uipage.transformer(uistate, previouspage)
        uistate.scrollToBottom = true
        local gamestate = uistate.core.gamestate
        if backup and gamestate.combatstate then
            log.info("Backing up.")
            gamestate.combatstate.previous = scarlettables.copy(
                gamestate.combatstate
            )
        end
        gamepage.transformer(gamestate)
        if transformer then
            transformer(uistate, previouspage)
        end
    end
    function uipage.drawgui(uistate)
        drawgui(uistate, uipage)
    end
    if gamepage.combatname then
        uipage.combatname = "ui;" .. gamepage.combatname
    end
    return uipage
end

--[[
-- Given a list of characters, returns a table of strings that displays them in a compact
-- manner.
--
-- The table is indexed on the index of that character in the provided list. Note this is 
-- *not* necessarily a sequence. There will be gaps in the indices if one or more characters
-- have been defeated
--
-- combatstate: The current state of combat
-- characters: The characters to display
-- prefix: A prefix to the selection key, if any
--
-- return: A list of strings representing a paragraph displaying all the characters in a compact
-- manner
-- ]]
function M.drawcharacters(combatstate, characters, prefix)
    local characterlabels = combatoperations.buildcharacterlabels(
        characters,
        prefix,
        combatstate.concise
    )
    local characterstrings = {}
    for i, character in ipairs(characters) do
        if not combatstateops.isdefeated(combatstate, character.name) then
            characterstrings[i] = characterlabels[character.name]
        else
            characterstrings[i] = ""
        end
    end
    return characterstrings
end

return M
