--[[
-- Copyright 2019 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
-- Implements Anklyana's powers. Anklyana is all about --dat bass-- defense.
-- She gets an extra point of energy per stat, and all of her powers revolve
-- around reducing the damage dealt to/by characters.
--]]

local attacks = require 'combat/attack'
local combatoperations = require 'combat/combatstate'
local F = require 'f-strings/F'
local log = require 'log'
local pages = require 'pages'
local scarletmoonpowers = require 'data/scarletmoonpowers'
local scarlettables = require 'scarlettables'
local statuses = require 'data/statuses'
local statusops = require 'combat/statuses'
local targeting = require 'combat/target'
local powerutils = require 'data/powerutils'

local M = {}

local function resolvetarget(targetkey, combatstate)
    if LOVE then
        if string.sub(targetkey, 1, 1) == "f" then
            targetkey = string.sub(targetkey, 2)
            return combatstate.allies[tonumber(targetkey)]
        else
            return combatstate.enemies[tonumber(targetkey)]
        end
    else
        return combatstate.characters[tonumber(targetkey)]
    end
end

------- Bubblesaurus -------
--[[
-- Increases the target's shield stat by *power*.
--]]
local function invokebubblesaurus(combatstate, previouspage, targetkey)
    local target = resolvetarget(targetkey, combatstate)
    local targetable = target
        and not combatoperations.isdefeated(combatstate, target.name)
    return pages.page(
        nil,
        function() if not targetable then return previouspage end
        end,
        nil,
        function(combatstate1)
            if not targetable then
                return 
            end
            local user = combatoperations.activecharacter(combatstate1)
            local power = attacks.statistic(user, "power")
            user.statistics.energy = user.statistics.energy - power
            local shield = target.secondarystatistics.shield
            local currentshield = shield.bubblesaurus or 0
            shield.bubblesaurus = currentshield + power
            if currentshield == 0 then
                table.insert(
                    combatstate1.combatlog,
                    F"{user.combatname} wraps {target.combatname} in a "
                        .. "green bubble!"
                )
            else
                table.insert(
                    combatstate1.combatlog,
                    F"{user.combatname} thickens the bubble around {target.combatname}!"
                )
            end
            powerutils.inflictcost(combatstate, M.bubblesaurus)
        end
    )
end

M.bubblesaurus = targeting.target(invokebubblesaurus, 1)
M.bubblesaurus.cost = 0
M.bubblesaurus.damage = 0
M.bubblesaurus.targets = combatoperations.activeallies
M.bubblesaurus.numtargets = 1
M.bubblesaurus.stats = {"power"}
function M.bubblesaurus.description(character)
    local power = attacks.statistic(character, "power")
    return F"Increases the target's shield by *{power}* points, at "
        .. F" the cost of *{power}* energy. Shield and cost increase with "
        .. "power."
end

---  Wallsaurus
--[[
-- Wallsaurus reduces a character's attack by power/4 and
-- increases their defense by power/4.
--]]

local function erectwall(user, target, combatstate)
    local power = attacks.statistic(user, "power")
    local buff = math.max(1, math.floor(power / 4))
    statusops.addstatus(target, statuses.wall, 1, user, buff, combatstate)
    table.insert(
        combatstate.combatlog,
        F"{user.combatname} erects a glowing green wall around "
            .. F"{target.combatname}!"
    )
end

local function applywallsaurus(combatstate, previoupage, targetkey)
    local target = resolvetarget(targetkey, combatstate)
    local targetable = target
        and not combatoperations.isdefeated(combatstate, target.name)
    if not targetable then
        return previouspage
    end
    local user = combatoperations.activecharacter(combatstate)
    return pages.page(
        nil,
        nil,
        nil,
        function(combatstate1) 
            erectwall(user, target, combatstate1)
        end
    )
end

--[[
-- Reduces damage dealt to and by a target by power/4 for 1 turn.
--]]
M.wallsaurus = targeting.target(applywallsaurus)

M.wallsaurus.cost = 1
M.wallsaurus.damage = 0
M.wallsaurus.targets = combatoperations.activecharacters
M.wallsaurus.numtargets = 1
M.wallsaurus.stats = {"power"}
function M.wallsaurus.description(character)
    local reduction = math.max(1, math.floor(character.statistics.power/4))
    return F"The target receives a *{reduction}* bonus to defense, and "
        .. F"an equal penalty to attack for 1 turn."
end

M.armorsaurus = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatoperations.activecharacter(combatstate)
        local combatname = character.combatname
        local power = attacks.statistic(character, "power")
        local buff = math.max(1, math.floor(power / 4))
        table.insert(combatstate.combatlog, F"{combatname}' shield thickens!")
        statusops.addstatus(
            character,
            statuses.armor,
            power,
            character,
            buff,
            combatstate
        )
    end
)

M.armorsaurus.cost = 5
M.armorsaurus.damage = 0
M.armorsaurus.targets = combatoperations.activecharacters
M.armorsaurus.numtargets = 0
M.armorsaurus.stats = {"power"}
function M.armorsaurus.description(character)
    local power = attacks.statistic(character, "power")
    local armor = math.ceil(power/4)
    return "Anklyana's defense increases by *" .. armor .."* points for *" .. power .. "* turns."
end

--[[
-- Grants Anklyana a status that purges all other statuses from her at the 
-- start of her turn, at a cost equal to the duration of each status purged.
--]]
M.stasisaurus = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatoperations.activecharacter(combatstate)
        local combatname = character.combatname
        local power = attacks.statistic(character, "power")
        local buff = math.max(1, math.floor(power))
        table.insert(combatstate.combatlog, F"{combatname}'s shield goes still.")
        statusops.addstatus(
            character,
            statuses.stasis,
            power,
            character,
            buff,
            combatstate
        )
    end

)

M.stasisaurus.cost = 0
M.stasisaurus.damage = 0
M.stasisaurus.targets = combatoperations.activecharacters
M.stasisaurus.numtargets = 0
M.stasisaurus.stats = {"power"}
function M.stasisaurus.description(character)
    local power = attacks.statistic(character, "power")
    local armor = math.ceil(power/4)
    return "Anklyana is healed of the worst status inflicitng her each turn for *" 
        .. power .. "*  turns and takes damage equal to the duration of the "
        .. "cured status."
end

--[[
-- Empathysaurus allows Anklyana to transfer somebody else's worst status to her. 
-- Yes, this includes buffs. Yes, this means
-- she could snag some of Scarlet Moon's buffs and gain, say, the Shoot action, not
-- that it would do her much good.
--]]
local function transferstatuses(user, target, combatstate)
    local worststatus
    local worstbuffvalue = 10000
    for i, statusduration in ipairs(target.statuses) do
        local status = statusduration.status
        local buff = status.buff
        if type(buff) == "function" then
            buff = buff(target)
        end
        if buff < worstbuffvalue then
            worststatus = statusduration
            worstbuffvalue = buff
        end
    end
    if worststatus then
        local status = worststatus.status
        table.insert(
            combatstate.log,
            F"{status.name} is transferred to {user.name} from {target.name}!"
        )
        statusops.removestatus(target, status, combatstate)
        statusops.addstatus(
            user,
            status,
            math.ceil(worststatus.duration/2),
            user,
            status.buff,
            combatstate
        )
    end
end

local function empathy(combatstate, invalidtarget, targetkey)
    local target = resolvetarget(targetkey, combatstate)
    local targetable = target
        and not combatoperations.isdefeated(combatstate, target.name)
    if not targetable then
        return previouspage
    end
    local user = combatoperations.activecharacter(combatstate)
    return pages.page(
        nil,
        nil,
        nil,
        function(combatstate1)
            transferstatuses(user, target, combatstate1)
        end
    )
end

M.empathysaurus = targeting.target(empathy)

M.empathysaurus.cost = 10
M.empathysaurus.damage = 0
M.empathysaurus.targets = combatoperations.activecharacters
M.empathysaurus.numtargets = 0
M.empathysaurus.stats = {"power"}
function M.empathysaurus.description(character)
    local power = attacks.statistic(character, "power")
    return "Anklyana transfers the worst status currently inflicting the "
        .. "chosen character to her, and cuts its duration in half."
end

return M
