--[[
-- Copyright 2018 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]
--[[
-- This module defines Buzzsaw's powers. Buzzsaw is a brutal fighter whose powers
-- focus primarily around doing as much damage as possible to Scarlet Moon.
--]]

local M = {}
local ai = require 'combat/ai'
local attacking = require 'combat/attack'
local combatstateops = require 'combat/combatstate'
local log = require 'log'
local pages = require 'pages'
local powers = require 'combat/powers'
local status = require 'data/statuses'
local statusops = require 'combat/statuses'
local targeting = require 'combat/target'

local function hisher(character)
    return character.attributes.gender == "female" and "her" or "his"
end

local function himher(character)
    return character.attributes.gender == "female" and "her" or "him"
end

-- --------------------- Buzzsaw ------------------------------
-- [[
-- Buzzsaw slashes an opponent with her mohawk, doing damage based on her
-- strength and power stats. The target protects themselves with
-- speed.
-- ]]

local function buzzsawdamage(user, target, combatstate)
    local userpower = attacking.attackstat(user)
    local targetpower = attacking.defensestat(target)
    local damage = math.max(1, userpower - targetpower)
    local description = user.combatname .. " snaps " .. hisher(user)
            .. " head forward, slashing " .. target.combatname
            .. " with " .. hisher(user) .. " hair, for "
            .. damage .. " damage!"
    local originalretaliation = attacking.retaliation(target)
    local retaliation = attacking.retaliation(user, target, damage)
    if originalretaliation then
        description = description 
            .. target.combatname .. " retaliates for " .. retaliation 
            .. " damage!"
    end
    return damage, (originalretaliation and retaliation or nil), {description}
end

M.buzzsaw = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local activecharacter = combatstateops.activecharacter(combatstate)
        powers.applycost(activecharacter, M.buzzsaw)
        local targets = {ai.targetmaxenergy(combatstate, activecharacter)}
        attacking.damage(combatstate, buzzsawdamage, targets, activecharacter)
    end
)

M.buzzsaw.cost = 4
M.buzzsaw.numtargets = 1
M.buzzsaw.stats = {"strength", "power"}
M.buzzsaw.description = "Buzzsaw slashes her opponent with her "
    .. "stronger-than-steel mohawk. Buzzsaw's hair is so sharp, she "
    .. "can even cut through shields. She is still vulnerable to retaliation." 

-- AI metadata
M.buzzsaw.targets = combatstateops.activeopponents
M.buzzsaw.damage = powers.DAMAGES
M.buzzsaw.aiexecution = function() return M.buzzsaw end

-- --------------------- Chainwhip ------------------------------
-- [[
-- Chainwhip applies a buff to Buzzsaw that adds her speed to her attack.
-- ]]
M.chainwhip = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        local power = attacking.statistic(character, "power")
        statusops.addstatus(
            character,
            status.chainwhip,
            2*power,
            character,
            nil,
            combatstate
        )
        table.insert(
            combatstate.combatlog,
            "Buzzsaw whips her chain above her head!"
        )
        powers.applycost(character, M.chainwhip)
    end
)

M.chainwhip.cost = 2
M.chainwhip.numtargets = 0
M.chainwhip.stats = {"speed", "power"}
M.chainwhip.description = "Buzzsaw starts attacking with her chain, adding "
    .. "her speed to her attack."

-- AI metadata
M.chainwhip.targets = combatstateops.activeallies
M.chainwhip.damage = powers.NEITHER

function M.chainwhip.aiexecution()
    return M.chainwhip
end

---------------------- Chainpull -------------------------------------------
--[[
-- Chainpull inflicts Grabbed on the target. A Grabbed target has their attack
-- and defense by Buzzsaw's strength/5, and heals them of Away at the start of 
-- their turn.
-- ]]

local function besttarget(speedstrengthdiff, targetstats)
    return not speedstrengthdiff
        or (targetstats.speed - targetstats.strength > speedstrengthdiff)
end

local function inflictchainpull(combatstate)
    local user = combatstateops.activecharacter(combatstate)
    local target
    local speedstrengthdiff
    local targets = combatstateops.opponents(combatstate, user)
    for _, possibletarget in pairs(targets) do
        local targetstats = possibletarget.statistics
        if besttarget(speedstrengthdiff, targetstats) then
            speedstrengthdiff = targetstats.strength - targetstats.speed
            target = possibletarget
        end
    end
    local description = user.combatname .. " wraps " .. hisher(user)
        .. " chain around " .. target.combatname .. ", holding " ..
        himher(target) .. " fast."
    table.insert(combatstate.combatlog, description)
    local userstrength = attacking.statistic(user, "strength")
    local targetstrength = attacking.statistic(target, "strength")
    local duration = math.max(2, userstrength - targetstrength)
    statusops.addstatus(target, status.grabbed, duration, user, nil, combatstate)
end

M.chainpull = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        inflictchainpull(combatstate)
        local activecharacter = combatstateops.activecharacter(combatstate)
        powers.applycost(activecharacter, M.chainpull)
    end
)

M.chainpull.cost = 2
M.chainpull.numtargets = 1
M.chainpull.stats = {"strength"}
M.chainpull.damage = powers.NEITHER
M.chainpull.targets = combatstateops.activeopponents
M.chainpull.description = "Buzzsaw wraps her target in chains, slowing her "
    .. "target down by max(1, buzzsaw.strength - target.strength). Duration is "
    .. "max(2, buzzsaw.strength - target.strength). Increases Buzzsaw's "
    .. "attrition by 2."
M.chainpull.aiexecution = function() return M.chainpull end

------------------------ Soaring Leap -------------------------------------
--[[
-- A buff that renders Buzzsaw immune to Away.
--]]

M.soaringleap = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local activecharacter = combatstateops.activecharacter(combatstate)
        local duration = 2 * attacking.statistic(activecharacter, "power")
        statusops.addstatus(
            activecharacter,
            status.leap,
            duration,
            activecharacter,
            nil,
            combatstate
        )
        local msg = activecharacter.combatname .. " flexes " 
            .. hisher(activecharacter) .. " legs and grins maniacally."
        table.insert(combatstate.combatlog, msg)
        powers.applycost(activecharacter, M.soaringleap)
    end
)

M.soaringleap.cost = 4
M.soaringleap.numtargets = 0
M.soaringleap.stats = {"strength"}
M.soaringleap.damage = powers.NEITHER
M.soaringleap.targets = combatstateops.activeallies
M.soaringleap.description = "Buzzsaw pumps power into her legs, allowing her "
    .. "to jump far distances on demand. Renders her immune to Away."

function M.soaringleap.aiexecution()
    return M.soaringleap
end


return M
