--[[
-- Copyright 2020 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]
--[[
-- This module defines Leopard's powers. Leopard is a perfectly balanced,
-- and brutal fighter. She primarily has straight damage attacks, but does
-- have one buff that grants her retaliation, and a debuff that moves
-- all of the target's stats into a single stat each turn.
--]]

local M = {}

local attacks = require 'combat/attack'
local combatstateops = require 'combat/combatstate'
local combatutils = require 'combat/combatutils'
local F = require 'f-strings/F'
local log = require 'log'
local pages = require 'pages'
local powers = require 'combat/powers'
local powerutils = require 'data/powerutils'
local scarlettables = require 'scarlettables'
local statuses = require 'data/statuses'
local statusops = require 'combat/statuses'

local lastdistortionindex = 1
--[[
-- Returns a target for distortion.
-- Distortion follows a very basic formula: It goes down
-- the list of targets in order, skipping over any who are already distorted.
--
-- Note that this does modify some state outside of this function
-- (lastdistortionindex), but that's fine. It doesn't really matter where we
-- start counting for a possible target, only that we pick a target who isn't
-- inflicted, and it isn't always the same target.
--
-- opponents: The list of possible targets for distortion
--
-- returns The target to hit with distortion
--]]
local function distortiontarget(opponents)
    if lastdistortionindex > #opponents then
        lastdistortionindex = 1
    end
    local lasttarget = opponents[lastdistortionindex]
    while statusops.statusduration(lasttarget, statuses.distortion) do
        lastdistortionindex = lastdistortionindex + 1
        if lastdistortionindex > #opponents then
            return opponents[lastdistortionindex]
        end
        lasttarget = opponents[lastdistortionindex]
    end
    return lasttarget
end

--[[
-- Distortion inflicts the distortion status on a character.
-- A character inflicted with distortion sees all their stats
-- transferred to one stat each turn.
--]]
M.distortion = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local activecharacter = combatstateops.activecharacter(combatstate)
        local name = activecharacter.combatname
        local opponents = combatstateops.opponents(combatstate, activecharacter)
        local target = distortiontarget(opponents)
        table.insert(
            combatstate.combatlog,
            F"{name} whips out a blowgun and fires it at {target.combatname}!"
        )
        local duration = attacks.statistic(activecharacter, "power")
        local description = F"{target.combatname}'s powers churn!"
        statusops.addstatus(
            target,
            statuses.distortion,
            duration, 
            user,
            combatstate
        )
        powerutils.inflictcost(combatstate, M.distortion)
    end
)

M.distortion.cost = 2
M.distortion.numtargets = 1
M.distortion.stats = {"power"}
M.distortion.damage = powers.NEITHER
M.distortion.targets = combatstateops.activeopponents
M.distortion.description = "Leopard inflicts Distortion on a target."
.. " A distorted target sees all their stat points transferred into a"
.. " different stat each turn."
M.distortion.aiexecution = function() return M.distortion end

--[[
-- A power where Leopard launches a massive kick that hits all opponents who
-- isn't inflicted with Away.
--]]
M.roundhousekick = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        table.insert(
            combatstate.combatlog,
            F"{character.combatname} unleashes a brutal roundhouse kick!"
        )
        local possibletargets = combatstateops.activeopponents(combatstate)
        local targets = {}
        for _, target in ipairs(possibletargets) do
            if not statusops.statusduration(target, statuses.away) then
                table.insert(targets, target)
                table.insert(
                    combatstate.combatlog, 
                    F"{character.combatname} kicks {target.combatname}!"
                )
            else
                table.insert(
                    combatstate.combatlog,
                    F"{target.combatname} is too far away to kick!"
                )
            end
        end
        attacks.damage(combatstate, nil, targets, character)
        powerutils.inflictcost(combatstate, M.roundhousekick)
    end
)

M.roundhousekick.cost = 10
M.roundhousekick.numtargets = powers.ALL_OPPONENTS
M.roundhousekick.targets = combatstateops.activeopponents
M.roundhousekick.stats = {"power"}
M.roundhousekick.damage = powers.DAMAGES
M.roundhousekick.description = "Leopard kicks every enemy who isn't Away."
M.roundhousekick.aiexecution = function() return M.roundhousekick end

--[[
-- Leopard will throw the opponent with the highest attack stat.
--]]
local function pickprojectile(opponents)
    local highestattack = opponents[1]
    for _, opponent in ipairs(opponents) do
        local higheststat = attacks.attackstat(highestattack)
        if higheststat < attacks.attackstat(opponent) then
            highestattack = opponent
        end
    end
    return highestattack
end

-- [[
-- Leopard prefers throwing at a target with retaliation. If there is no
-- target with the highest retaliation, Leopard will instead pick the target with the
-- lowest defense
--
-- opponents: The list of possible targets
-- projectile: The character being thrown, used so that Leopard doesn't
-- try to throw someone at themselves

-- return The character to throw the projectile at, or nil if there is 
--  no valid target (i.e. there is only one opponent left, and that's 
--  the projectile)
-- ]]
local function picktarget(opponents, projectile)
    local retaliatingopponent
    local mostretaliation = 0
    for _, opponent in pairs(opponents) do
        if opponent.name ~= projectile.name then
            local retaliation = attacks.retaliation(opponent)
            if mostretaliation < retaliation and projectile ~= opponent then
                retaliatingopponent = opponent
                mostretaliation = retaliation
            end
        end
    end
    if retaliatingopponent then
        return retaliatingopponent
    end

    local vulnerableopponent
    local lowestdefense = -1
    for _, opponent in pairs(opponents) do
        if opponent.name ~= projectile.name then
            local defense = attacks.defensestat(opponent)
            if lowestdefense < 0 or defense < lowestdefense then
                vulnerableopponent = opponent
                lowestdefense = defense
            end
        end
    end
    return vulnerableopponent
end

--[[
-- Leopard throws one opponent at another. Note that while Leopard is
-- damaging the thrown opponent, the thrown opponent is the one damaging
-- the targeted opponent, and thus damage is done based on the projectile's
-- attack score, and any retaliation will be directed to the thrown character.
--]]
M.throw = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        local opponents = combatstateops.activeopponents(combatstate)
        local projectile = pickprojectile(opponents)
        local projectilename = projectile.combatname
        local target = picktarget(opponents, projectile)
        local msg
        if target then
            local targetname = target.combatname
            msg = F"{character.combatname} throws {projectilename} at {targetname}!"
        else
            msg = F"{character.combatname} throws {projectilename}!"
        end
        table.insert(combatstate.combatlog, msg)
        attacks.damage(combatstate, nil, {projectile}, character)
        if target then
            attacks.damage(combatstate, nil, {target}, projectile)
        end
    end
)

M.throw.cost = 7
M.throw.numtargets = 2
M.throw.stats = {"strength"}
M.throw.damage = powers.DAMAGES
M.throw.targets = combatstateops.activeopponents
M.throw.description = "Leopard throws one target at another. "
.. "Effectively, Leopard is attacking the projectile, while the projectile "
.. "is attacking the other target."
M.throw.aiexecution = function() return M.throw end

--[[
-- Inflicts Leopard with the Counter status, granting her retaliation
-- equal to double her attack.
--]]
M.judo = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        table.insert(
            combatstate.combatlog,
            F"{character.combatname} takes a judo stance, ready for attack."
        )
        local duration = attacks.statistic(character, "power")
        statusops.addstatus(character, statuses.counter, duration, combatstate)
    end
)

M.judo.cost = 10
M.judo.numtargets = 0
M.judo.stats = {"power"}
M.judo.damage = powers.NEITHER
M.judo.targets = combatstateops.activeallies
function M.judo.description(character)
    local retaliation = 2 * attacks.attackstat(character)
    return "Leopard enters a judo stance, allowing her to retaliate. "
        .. " She gains *" .. retaliation .. "* retaliation."
    end
M.judo.aiexecution = function() return M.judo end

return M
