--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
-- Implements Prometheus' powers. Prometheus manipulates fire, and has two
-- modes: warm and aflame. When warm, she uses her fiery based powers, which are
-- almost exclusively offensive, though she does have a couple of buffs. When
-- aflame, all of her attacks do double damage, and she doesn't take damage
-- with every action like others do.
--]]

local M = {}
local attacks = require 'combat/attack'
local combatstateops = require './combat/combatstate'
local log = require './log'
local pages = require './pages'
local powers = require 'combat/powers'
local powerutils = require './data/powerutils'
local statuses = require './data/statuses'
local statusops = require 'combat/statuses'

local function fireballdamage(user, target, combatstate)
    local function damageformula(user, defender)
        local attack = math.floor(1.5 * attacks.attackstat(user, "power"))
        local defense = attacks.defensestat(defender)
        return math.max(2, attack - defense)
    end
    local function damage(attacker, defender, combatstate)
        local damage, retaliation = attacks.defaultdamage(
            attacker,
            defender,
            combatstate,
            damageformula
        )
        local description = attacker.combatname
            .. " flings a fireball at "
            .. defender.combatname .. " for " .. damage .. " damage!"
        return damage, retaliation, {description}
    end
    attacks.damage(combatstate, damage, {target}, user)
end

local function inflictfireball(attacker, defender, combatstate)
    local iswarm = statusops.statusduration(defender, statuses.warm) 
        or statusops.statusduration(defender, statuses.aflame)
    local status = iswarm and statuses.aflame or statuses.warm
    local attackpower = attacks.statistic(attacker, "power")
    local defenderpower = attacks.statistic(defender, "power")
    local duration = math.max(2, attackpower - defenderpower)
    local description = iswarm
        and defender.combatname .. " burns."
        or defender.combatname .. " feels warm."
    table.insert(combatstate.combatlog, description)
    statusops.addstatus(defender, status, duration, attacker, nil, combatstate)
end

local function fireballfreeze(user, target, combatstate)
    local status = statuses.frozen
    local wet = statusops.statusduration(target, statuses.wet)
    if wet then
        local potency = math.floor(wet.duration/2)
        statusops.addstatus(target, status, wet.duration, user, potency, combatstate)
        statusops.removestatus(target, statuses.wet, combatstate)
        table.insert(combatstate.combatlog, target.combatname .. " freezes!")
    end
end

--[[
-- Given the current state of combat, and the user of fireball, returns a
-- target to freeze as a part of using fireball. The game prefers to freeze
-- Poseidon (since it's a buff for him), but can only do so if he's wet.
-- Otherwise, the Prometheus freezes the first wet opponent she can find. If
-- there is nobody to freeze, this function returns nil.
--]]
local function findfreezetarget(combatstate, user)
    for _, character in ipairs(combatstateops.allies(combatstate, user)) do
        local canfreezeposeidon =
            statusops.statusduration(character, statuses.wet)
            and not statusops.statusduration(character, statuses.frozen)
            and string.find(character.name, "poseidon")
        if canfreezeposeidon then
            return character
        end
    end
    local opponents = combatstateops.opponents(combatstate, user)
    for _, opponent in pairs(opponents) do
        if statusops.statusduration(opponent, statuses.wet) then
            return opponent
        end
    end
end

local function fireball(combatstate)
    local user = combatstateops.activecharacter(combatstate)
    local targets = combatstateops.activeopponents(combatstate)
    local warmtargets = {}
    for _, target in ipairs(targets) do
        if statusops.statusduration(target, statuses.warm) then
            table.insert(warmtargets, target)
        end
    end
    local target = powerutils.median(
        next(warmtargets) and warmtargets or targets,
        combatstate.round
    )
    if not target then
        return
    end
    local freezetarget = findfreezetarget(combatstate, user)
    fireballdamage(user, target, combatstate)
    inflictfireball(user, target, combatstate)
    if freezetarget then
        fireballfreeze(attacker, freezetarget, combatstate)
    end
end

M.fireball = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        fireball(combatstate)
        powerutils.inflictcost(combatstate, M.fireball)
    end
)

M.fireball.damage = 1
M.fireball.targets = combatstateops.activeopponents
M.fireball.cost = 4
M.fireball.numtargets = 1
M.fireball.stats = {"power", "speed"}
M.fireball.description = "Prometheus does 1.5 * power + speed/2 "
    .. "damage to one target and inflicts warm. If target is already warm, "
    .. "inflicts aflame instead. Aflame deals damage based on the status' "
    .. " current duration. Freezes one wet target, if any."
M.fireball.aiexecution = function() return M.fireball end

M.twinserpents = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        local duration = attacks.statistic(character, "power")
        local name = character.combatname
        local description = "Fiery serpents twine around " .. name .. "."
        table.insert(combatstate.combatlog, description)
        statusops.addstatus(character, statuses.twinserpents, duration, combatstate)
        statusops.addstatus(character, statuses.warm, duration, combatstate)
        powerutils.inflictcost(combatstate, M.twinserpents)
    end
)

M.twinserpents.damage = 0
M.twinserpents.targets = combatstateops.activeallies
M.twinserpents.cost = 4
M.twinserpents.numtargets = 0
M.twinserpents.stats = {"power"}
M.twinserpents.description = "Prometheus twines two fiery serpents around "
    .. "herself, granting her an extra Power use for power turns. Also inflicts"
    .. " warm on herself."
M.twinserpents.aiexecution = function() return M.twinserpents end

-- Flamey is a summoning spell that allows Prometheus to use half her current
-- energy to effectively create a copy of herself.
M.flamey = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        local summon = powerutils.summon(combatstate, character, "Flamey", "F")
        if summon then
            local combatname = character.combatname
            local description = combatname 
                .. " swirls a ball of flames into a fiery serpent!"
            table.insert(combatstate.combatlog, description)
            powers.applycost(character, M.flamey)
        end
    end
)
M.flamey.damage = 0
M.flamey.targets = combatstateops.activeallies
M.flamey.cost = 15
M.flamey.numtargets = 0
M.flamey.stats = {"strength", "speed", "power"}
M.flamey.description = "Prometheus weaves her flames into an autonomous fiery serpent with all "
    .. " her strength and powers. However, she does have to split her energy between herself and "
    .. " her serpent. Costs *15* energy."
M.flamey.aiexecution = function() return M.flamey end

return M
