--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
-- ]]
--
-- [[
-- This module defines Succubus' powers. The bulk of Succubus' powers revolve
-- around redirection and debuffs. She only has a single attack power: Crop whip
-- which does damage and inflicts the Spanked ailment.
--
-- Most of her powers instead revolve around controlling or otherwise
-- inhibiting her opponents. However, these powers have a critical weakness:
-- they only really work when Succubus' power is greater than her opponent's.
-- Many of them are also unable to be used if Succubus has been spanked
-- recently.
-- ]]

local attacks = require 'combat/attack'
local combatstateops = require 'combat/combatstate'
local F = require 'f-strings/F'
local log = require 'log'
local pages = require 'pages'
local powers = require 'combat/powers'
local powerutils = require 'data/powerutils'
local scarletchoices = require 'data/scarletchoices'
local statuses = require 'data/statuses'
local statusops = require 'combat/statuses'

local M = {}

local draindescription = "Otherwise reduce target's power by 1."


-- [[
-- Given a user, a target, and a description, inflicts the drain stat on the
-- target for a duration equal to the user's power.
--
-- user: The character inflicting drain
-- target: The target of drain
-- description: A string describing previous actions by the user
--
-- return The drain status, a duration of equal to the user's power, and a
-- list containing the passed description and a new string describing the
-- infliction of drain
-- ]]
local function inflictdrain(user, target, description, combatstate)
    local resist = target.combatname
        .. " sways and shakes "
        .. powerutils.hisher(target)
        .. " head."
    table.insert(combatstate.combatlog, resist)
    local userpower = attacks.statistic(user, "power")
    statusops.addstatus(target, statuses.drain, userpower, combatstate)
end

--[[
-- Given a list of possible targets, returns all targets whose power is lower
-- than the active character's power. If none of the targets have a lower
-- power, then this function returns the original list.
--
-- combatstate: The current state of combat
-- targets: The list of targets to filter
--
-- return The sublist of targets whose power is lower than power, or the
-- original list if there are no such targets
--]]
local function lowpower(combatstate, targets)
    local character = combatstateops.activecharacter(combatstate)
    local power = attacks.statistic(character, "power")
    local lowpowertargets = {}
    for _, target in ipairs(targets) do
        if attacks.statistic(target, "power") < power then
            table.insert(lowpowertargets, target)
        end
    end
    return #lowpowertargets > 0 and lowpowertargets or targets
end

-------------- Stop Hitting Yourself ----------------------------
--[[
-- Stop Hitting Yourself gives Succubus retaliation
-- based on her power * 5.
--]]

M.stophittingyourself = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        local retaliation = attacks.statistic(character, "power") * 5
        local retaliationmodifiers = character.secondarystatistics.retaliation
        retaliationmodifiers[M.stophittingyourself.name] = retaliation
        local msg = activecharacter.combatname 
            .. "smirked. \"You realy shouldn't hit yourself. That can't be "
            .. "healthy.\""
        table.insert(combatstate.combatlog, msg)
        powerutils.inflictcost(combatstate, M.stophittingyourself)
    end
)
M.stophittingyourself.damage=powers.NEITHER
M.stophittingyourself.targets=combatstateops.activeallies
M.stophittingyourself.cost = 1
M.stophittingyourself.numtargets=0
M.stophittingyourself.stats={"power"}
function M.stophittingyourself.description(character)
    local retaliation = attacks.statistic(character, "power") * 5
    return "Succubus gains *" .. retaliation .. "* points of retaliation."
end
M.stophittingyourself.aiexecution = function() return M.stophittingyourself end

-------------- Corner Time --------------------------------------
--[[
-- Inflicts away on the target, if the target has a lower power, for
-- a duration equal to the userpower - targetpower.
--]]
--

--[[
-- Inflicts away on a target if their power is lower than Succubus',
-- and drain on them if it isn't.
--]]
local function inflictcornertime(combatstate)
    local description = '"Go stand in the corner '
        .. target.combatname ..
        '. And think about what you\'ve done," said Succubus.'
    local targets = lowpower(combatstate, M.cornertime.targets(combatstate))
    local besttarget
    local higheststrength
    for _, target in ipairs(targets) do
        local targetstrength = attacks.statistic(target, "strength")
        if not higheststrength or targetstrength > higheststrength then
            higheststrength = targetstrength
            besttarget = target
        end
    end
    local user = combatstateops.activecharacter(combatstate)
    local userpower= attacks.statistic(user, "power") 
    local targetpower = attacks.statistic(besttarget, "power")
    local duration = userpower - targetpower
    if 0 < duration then
        statusops.addstatus(besttarget, statuses.away, duration, combatstate)
        table.insert(
            combatstate.combatlog,
            target.combatname .. " hangs " .. hisher(target) .. " head "
            .. "in shame, and walks into the corner."
        )
    else
        local duration = attacks.statistic(user, "power")
        statusops.addstatus(target, statuses.drain, duration, combatstate)
        table.insert(
            combatstate.combatlog,
            target.combatname .. " shakes " .. hisher(target) .. " head."
        )
    end
end

M.cornertime = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        inflictcornertime(combatstate)
        powerutils.inflictcost(combatstate, M.cornertime)
    end
)

M.cornertime.damage=powers.NEITHER
M.cornertime.targets=combatstateops.activeopponents
M.cornertime.cost = 1
M.cornertime.numtargets=1
M.cornertime.stats={"power"}
M.cornertime.description=
    "If Succubus.power > target.power, inflict Away." .. draindescription
M.cornertime.aiexecution = function() return M.cornertime end

------------- Over My Knee --------------------------------------
--[[
-- Inflicts the bent over status on a target. Bent over drops all actions to
-- zero.
--]]

--[[
-- Inflicts overmyknee on a target.
--]]
local function inflictovermyknee(combatstate)
    local targets = lowpower(combatstate, M.overmyknee.targets(combatstate))
    local besttarget
    local targetname
    for _, target in pairs(targets) do
        local targetpower = attacks.statistic(target, "power")
        local bettertarget = not besttarget
            or targetpower < attacks.statistic(besttarget, "power")
        if bettertarget then
            targetname = target.name
            besttarget = target
        end
    end

    local user = combatstateops.activecharacter(combatstate)
    local description = '"Get over my knee, '
        .. besttarget.combatname .. ","
        .. ' you little brat," said Succubus.'
    local userpower = attacks.statistic(user, "power")
    if attacks.statistic(besttarget, "power") < userpower then
        local submit = besttarget.combatname
            .. " approaches Succubus and bends over her knee, pushing "
            .. powerutils.hisher(besttarget)
            .. " bottom up into the air."
        description = description .. " " .. submit
        table.insert(combatstate.combatlog, description)
        statusops.addstatus(besttarget, statuses.grabbed, 2, user, nil, combatstate)
        statusops.addstatus(besttarget, statuses.otk, 2, user, nil, combatstate)
    else
        table.insert(combatstate.combatlog, description)
        inflictdrain(user, besttarget, description, combatstate)
    end
end

M.overmyknee = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        inflictovermyknee(combatstate)
        powerutils.inflictcost(combatstate, M.overmyknee)
    end
)

M.overmyknee.damage=powers.NEITHER
M.overmyknee.targets=combatstateops.activeopponents
M.overmyknee.cost = 1
M.overmyknee.numtargets=1
M.overmyknee.stats={"power"}
M.overmyknee.description=
    "If Succubus.power > target.power, target bends over "
        .. "Succubus' knee for one round. Target cannot do anything, "
        .. "Succubus' spankings more effective. Not useable if spanked."
        .. draindescription
M.overmyknee.aiexecution = function() return M.overmyknee end

------------ Crop Whip ----------------------------------------------
--[[
-- With Crop whip, Succubus spanks a target with her riding crop, doing
-- damage and inflicting the Spanked status.
--]]

local function cropdamage(user, target, combatstate)
    local function whipdamage(attacker, attackee, combatstate1)
        local damage, retaliation = attacks.defaultdamage(
            attacker,
            attackee,
            combatstate1
        )
        local otk = statusops.statusduration(target, statuses.otk)
        local description = user.combatname
            .. (otk and (" holds " .. target.combatname ..  " down and ") 
                or " grabs " .. target.combatname .. " around the arm and ")
            .. " whips the "
            .. powerutils.heroheroine(target) .. "'s"
            .. powerutils.lowerrisque(target, " exposed ", " lightly clad ")
            .. " buttocks with " .. powerutils.hisher(user) .. " crop "
            .. (damage > 0 and (damage .. " times!") or "several times!")
        if otk then
            damage = 2 * damage
            retaliation = retaliation and math.floor(retaliation / 2)
            description = description .. " "
                .. target.combatname .. " bucks against "
                .. user.combatname  .. "'s"
                .. " knee, kicking and yelping while the crop leaves long thin "
                .. " welts on " .. powerutils.hisher(target)
                .. " jiggling bottom."
        else
            description = description .. " "
                .. target.combatname .. " scampers away, clutching at "
                .. powerutils.hisher(target) .. " stinging bottom."
        end
        return damage, retaliation, {description}
    end
    attacks.damage(combatstate, whipdamage, {target}, user)
end

local function cropwhip(combatstate)
    local user = combatstateops.activecharacter(combatstate)
    local targets = lowpower(combatstate, M.cropwhip.targets(combatstate))
    local otktargets = {}
    for _, target in ipairs(targets) do
        if statusops.statusduration(target, statuses.otk) then
            table.insert(otktargets, target)
        end
    end
    if next(otktargets) then targets = otktargets end
    local besttarget
    for _, target in ipairs(targets) do
        local targetenergy = attacks.statistic(target, "energy")
        if not besttarget then
            besttarget = target
        elseif targetenergy < attacks.statistic(besttarget, "energy") then
            besttarget = target
        end
    end
    cropdamage(user, besttarget, combatstate)
    statusops.addstatus(besttarget, statuses.spanked, 1, user, nil, combatstate)
end

M.cropwhip = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        cropwhip(combatstate)
        powerutils.inflictcost(combatstate, M.cropwhip)
    end
)

M.cropwhip.damage=powers.DAMAGES
M.cropwhip.targets=combatstateops.activeopponents
M.cropwhip.cost = 1
M.cropwhip.numtargets=1
M.cropwhip.stats={"strength", "speed", "power"}
M.cropwhip.description=
    "Succubus whips her opponent with her riding crop, "
        .. "doing standard "
        .. " damage and inflicting Spanked for *power* turns. Spanked "
        .. " reduces the duration of the target's best buff."
        .. " This power is twice as effective if "
        .. " target is bent over, or otk."
M.cropwhip.aiexecution = function() return M.cropwhip end

----------- Succuspank --------------------------------------------
--[[
-- Succuspank is like cropwhip, except it reduces the target's best
-- buff based on Succubus' power stat rather than her relative strength.
--]]

local function inflictsuccuspank(user, target, combatstate)
    local status = statuses.spanked
    local duration = attacks.statistic(user, "power")
    local description = F("{user.combatname}"
        .. " swats {target.combatname}'s upthrust, "
        .. "{powerutils.lowerrisque(target, 'naked', 'bucking')} buttocks."
    )
    if statusops.statusduration(target, statuses.otk) then
        description = description
            .. " " .. target.combatname .. " yelps and bucks against "
            .. user.combatname .. "'s" .. " knee. "
            .. powerutils.HeShe(target) .. " kicks and flails while "
            .. powerutils.hisher(target) .. " bottom wobbles beneath the "
            .. "assault."
    end
    table.insert(combatstate.combatlog, description)
    statusops.addstatus(target, statuses.succuspanked, 2, user, combatstate)
end

local function succuspank(combatstate)
    local targets = lowpower(combatstate, M.succuspank.targets(combatstate))
    local otktargets = {}
    for _, target in ipairs(targets) do
        if statusops.statusduration(target, statuses.otk) then
            table.insert(otktargets, target)
        end
    end
    if next(otktargets) then
        targets = otktargets
    end
    local besttarget
    for _, target in ipairs(targets) do
        local targetstrength = attacks.statistic(target, "strength")
        local newbesttarget = not besttarget
            or targetstrength < attacks.statistic(besttarget, "strength")
        if newbesttarget then
            besttarget = target
        end
    end
    local user = combatstateops.activecharacter(combatstate)
    inflictsuccuspank(user, besttarget, combatstate)
end

M.succuspank = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        succuspank(combatstate)
        powerutils.inflictcost(combatstate, M.succuspank)
    end
)

M.succuspank.damage=powers.NEITHER
M.succuspank.targets=combatstateops.activeopponents
M.succuspank.cost = 1
M.succuspank.numtargets=1
M.succuspank.stats={"strength", "speed", "power"}
M.succuspank.description=
    "Succubus spanks her opponent,"
        .. " inflicting Spanked for 2*power turns. Spanked "
        .. " reduces the duration of the target's best buff by 2*power turns."
        .. " Twice as effective if target is "
        .. "bent over or otk."
M.succuspank.aiexecution = function() return M.succuspank end

return M
