# Table of Contents #

- [Frontend](#frontend)
    - [UI Pages](#ui-pages)

- [Backend](#backend)
    - [Story Pages](#story-pages)
    - [Combat Pages](#combat-pages)


# Frontend #

The frontend works with the UI State via UI pages. User input (translated
from the UI framework's notion of user input) serves as the keys into UI pages.
The body and transformer are how we "step into" the backend where we deal
with game pages and game state. The body is our vehicle through which we
surface feedback from game selections (i.e. the amount
of damage done to an opponent), while the transformer is how we we update the
game and UI states (i.e. by invoking the attack action's transformer on the
game state that the UI state wraps).

## drawgui ##

All UI pages are expected to have an additional operation `drawgui`. `drawgui`
is an operation that takes the current `uistate` and builds the GUI for that
page.

## UI Pages ##

We have the following types of UI pages:

* Story UI page - Pages dealing with displaying the story.
* Combat UI page - Pages dealing with combat.
* Uilities UI Page - Pages dealing with utilities (i.e. saving, loading,
    exiting, character status)

### UI Transformers ###

Page transformers are applied when first transitioned to. Furthermore,
the `transformer` is the _only_ place where state change is allowed. Since
UI pages are created dynamically (rather than statically like story pages),
this means that if a particular UI page needs to change some state based on
choices made by the player, it wraps the state change in the `transformer` of
the _next_ UI page.

As an illustrative example, suppose we have a very simple command
`(C)hange Difficulty`. When the user selects `C`, a `Change Difficulty Page`
may be created, and provides a list of three choices: `(E)asy`, `(M)edium`,
`(H)ard`. When the player selects `M`, `choices.M` will return a function
that when invoked returns a new [UI Story page](#ui-story-page) whose
transformer sets the difficulty to medium.

### UI Body ###

All UI bodies return a table of
[SUIT](http://suit.readthedocs.io/en/latest/index.html) widgits fully
describing the UI appropriate for the page. The widgits are structured
from left to right, top to bottom so that the invoker of this method need
only dumbly add them to the `suit`.

### Story UI Page ###

This corresponds to the mode where the user is interacting with the story.

#### Body ####

In order to ensure that text is displayed correctly even in the case of
side-effects from the transform page of previous automatics, the body function
will be a bit interesting. It will perform a deep copy of the state, and then
iterate through each page in the automatics chain, applying each story page's
transformer after extracting its text.

Story UI pages effectively make it look like story pages have no automatics,
by merging pages linked by automatic transitions into a single giant page.
With all side-effects applied when leaving that page (though the game text
still looks like the `do` operations are applied when they should be).

#### Automatics ####

None. Story UI pages collapse automatic chains into single
pages that contains all of the state modification commands and text, with the
choices displayed in the last page in the chain. This drastically simplifies
displaying the story text.

#### Choices ####

When the player makes a choice, then the `choices` function forwards to the game
page's choices function to get the next game page. Then, based on the type of
the game page, it builds and returns a UI page based on the type of
game page.

Currently, there are two kinds of pages that could be constructed:

* [Story](#story-ui-page)
* [Combat](#combat-ui-page)

In addition, there are also several utility choices:

* `(Up) Previous line`
* `(Down) Next line`
* `(Page Up) Previous Page`
* `(Page Down) Next Page`
* `(O)ptions`
* `(S)ave`
* `(L)oad`

Each builds the next appropriate page.

The `choices` function takes the UI state and returns a table that maps
Love KeyConstants to pages.

#### Transformer ####

The transformer passes the game state to the game page's
transformer, as well as update the game state's `currentPage` key to the
current page.

It also sets the `uiState.parser` key to the appropriate parser from
`uiState.allparsers` based on the type of page being returned.

#### Text ####

A function that given the gamestate, returns the resolved text of this
page. Used internally to extract the text of automatic pages without
inflicting unintentional side-effects.

### Combat UI Page ###

This page models command selection at the start of each round.

#### Body ####

The body consists of a table of widgits fully describing the combat UI, where
the widgits are ordered from left to right, top to bottom.

#### Automatics ####

None

#### Choices ####

The choices are an ordered table of all the commands available to the player
(i.e. attack, defend, power), and will transition to a selection page, where
the player can select targets/powers/etc. Note that the selection page's
body will need to close over the current UI structure.

#### Transformer ####

Standard UI transformer

#### Positions ####

This entry maps to a function that takes the `uistate` and returns a table
describing the positions of the characters in battle.

##### Character table structure #####

- `name` - The name to display for the character.
- `energy` - The character's [energy](./characters.md#energy).
- `maxenergy` - The character's [maximum energy](./characters.md#energy).
- `statuses` - A table containing the remaining duration of each status
    inflicted on the character, keyed on status name.


### Selection UI Page ###

A selection UI page is used whenever the character needs to make a reversible
selection in battle.

#### Body ####

Returns a table of widgits fully describing the UI structure. The bulk of the
structure should have been closed over as a part of constructing this page
by whatever UI page built this structure. This page will merely add on to
the end a description of the current state of the selection, and the list of
choices available.

##### Automatics ####

If the character cannot perform an action for any reason, then we automatically
transition to a `no-op` action.

#### Choices ####

The following choices are possible:

- `(#)/(F#) Select`
- `(#) Select`
- `(F#) Select`
- `Select from List`
- `(<==) Back`
- `(Enter) Done`

The first 4 will all return new selection pages with transformers that save
the current `uistate` as `uistate.previous` and then apply the selection.

`(<==) Back` returns a copy of `uistate.previous.currentPage`,
but with a different `transformer`. In the transformer, it
overwrites `uistate` with `uistate.previous`. In other words, it undoes
any changes made _while transitioning to this page_.

`(Enter) Done` needs to do a couple of different things:

1. Returns the next [combat UI page](#combat-ui-page) with a
transformer that fully applies the selection (i.e. building
a combat page), and iterates to the next character to select actions for.
This should happen if we are currently selecting targets.

2. Return a [selection ui page](#selection-ui-page) with a transformer that
selects the chosen power as the action to use.

Which one it does will depend on the type of selection page, and will be
decided when this page is built.

#### Positions ####

Returns the same thing as the `positions` key in
[Combat UI Page](#combat-ui-page), except that the table has an additional
key `targeted`

- `targeted` - `true` iff the character has been targeted by the current
    selection, `false` otherwise.

Implementation note: Since `nil` is false, we can use the same logic to
translate the output of this function into the UI as we do for the `positions`
key in [Combat UI Page](#combat-ui-page).

### Confirm Page ###

A very simple page that simply serves as confirmation before a round starts.

#### Body ####

Returns a widgit that displays the list of actions selected by the player, and
the commands available.

#### Automatics ####

None

#### Choices ####

* `(Enter) Confirm` - Returns a [Combat UI Page](#combat-ui-page) whose
transformer applies the combat round, and stores the results for the body
to return.

* `(<==) Back` - Does the same thing as the choice in the
    [Selection UI Page](#selection-ui-page).

#### Transformer ####

Does nothing extra beyond applying the selection from the previous command.

### Save/Load Pages ###

These pages handle saving and loading gamestate.

#### Body ####

For a save page, asks the user where they want to save. For a load page,
asks where they want to load. Both pages display a list of all current
savegames.

#### Automatics ####

None

#### Choices ####

* `Alphanumeric`: Returns a new save/load page whose choice functions return
    the characters that have been typed by the player so far.
* `(Enter) Save/Load`: Returns an page whose transformer saves/loads the
    game to/from the specified file. In the case of `save`, the game then
    invokes `back`. In the case of loading, returns a `story` page that
    displays `uistate.body`, while `automatics` and `choices` are standard
    for a `story` page.

#### Transformer ####

Sets `uistate.previous` but other than that does nothing special beyond what
was specified by the previous page.

### Exiting ###

#### Body ####

Asks for confirmation before exiting.

#### Automatics ####

None

#### Choices ####

* `(Enter) Confirm` - Builds an page with just a transformer, which closes
    the game
*  `(<==) Back` - Standard back action

#### Transformer ####

Standard UI transformer

### Character Status Selection ###

#### Body ####

Returns a function that returns the empty table.

#### Automatics ####

If the party is of size 1, immediately transition to
[Character Status](#character-status).

#### Choices ####

* `(F#) - Select Character` - Builds a
    [Character Status](#character-status) page to display the selected
    character.
* `(<==) - Back` - Standard transition back

#### Transformer ####

Standard UI transformer.

### Character Status ###

#### Body ####

Returns a table describing the character's status:

* `name` - The printed name of the character
* `gender` - The character's gender
* `powers` - The character's powers
* `equipment` - The character's equipment
* `inventory` - The character's inventory
* `primaryStatistics` - The character's primary statistics
* `secondaryStatistics` - The character's secondary statistics
* `appearance` - The character's appearance

#### Automatics ####

None


#### Choices ####

* `(<==) Back` - Standard UI back


#### Transformer ####

Standard UI transformer

### Options Page ###

The Options page handles setting the different options available to the
player.

#### Body ####

Returns a table of each option keyed on option name.

Each option is a table consisting of:

* `choices` - The list of possible settings for this option
* `setting` - The current setting of the option

#### Automatics ####

None

#### Choices ####

* `Select from List` - Builds an
    [Options Selection Page](#options-selection-page) whose choice
    function closes over the selected option, (and this option page).
* `(Enter) Done` - Builds a copy of the previous page, except with a
transformer that applies the new options
* `(<==) Back` - Standard UI back operation


#### Transformer ####

No-op

### Options Selection Page ###

#### Body ####

Identical to [Options Page](#options-page)

#### Choices ####

* `Select from List` - Builds a [Options Page](#options-page) that is
    a copy of the options page that triggered this one, except the transformer
    returned by the options page's `(Enter) Done` page also applies this
    option. Note that the new options page back operation is identical to
    the original. In other words, an option selection page's back operation
    does not remember any option selection pages.

* `(<==) Back` - Standard UI back operation

#### Automatics ####

None

#### Transformer ####

No-op

# Backend #

The backend is where all the fancy-pants processing takes place. Like the
frontend, there are two main components: state (`gamestate`) and pages, which
are just like `ui pages` in that they are blocks of functions/operations on
`gamestate` that are invoked contiguously.

## Story Pages ##

The bulk of the game will be spent processing these pages. Story pages are
models of the CYOA-style blocks of text and choices that the player interacts
with. In addition to the other keys available on a page
(body, choices, automatics, and transformer), story pages also have a key `name`
that is the name of this page.

### Defining Story Pages ###

The bulk of a game developer's time will be spent defining story pages. They
follow the following structure:

```yaml
<EVENT_NAME>:
  music: <OPTIONAL_NAME_OF_SONG_TO_PLAY>
  text: >
    Template text.
  automatic choices:
    predicate1: page_name1
    predicate2: page_name2
  choices:
    template text: page_name3
    template text: page_name4
  do:
    - operation1
    - operation2
```

- `name` - a unique name for the page.
- `music` - an optional parameter that specifies the name of the song to play when
    this page begins.
- `text` - an optional parameter that specifies the dynamic text that will be
    resolved into text to show to the user when the page plays.
- `automatic choices`- a map of predicates on gamestate to page names. When the
   page is played, the game will transition automatically to one of these
   pages if the predicate is satisfied. The predicates are evaluated in the
   order specified.
- `choices` - Explicit choices available to the player. The keys are dynamic
    text (just like that passed to `text`) displayed to the user. The values
    are the page names to transition to for each choice. The choices are
    displayed in the order specified.
- `do` - A list of operations that should be performed on the gamestate (for
    example, adding a keyword, or improving one of the player's statistics)
    when the page is _finished_.

For more information about story pages, see [Stories](./story/stories.md)

#### Note on `do` ####

It should be noted that `do` in the YAML and `transformer` in the page
model _are not the same thing_. The `transformer` is invoked when entering
an page, while `do` is invoked when leaving an page. In other words,
`do` operations will in fact be invoked as a part of the _next_ page's
transformer.

This is confusing, and unfortunate. However, it is done for a couple of reasons:

* By invoking the `transformer` before anything else, we don't have to modify
the current page based on choices made by the player. Instead, we can wrap
any state changes that need to be made in the next page's transformer as a
part of the `choices` function. This allows us to restrict side-effects to
the `transformer` on state, and each function in each page can be
self-contained. In other words, this allows pages to be loose, self-contained
tables of functions/operations, rather than tightly interwoven objects with
their own state.

* By invoking `do` when leaving an page, rather than entering it, we don't have
to do anything different when loading a game vs playing the next page. If
we invoked the `do` commands upon _entering_ a state, then we'd need to
make sure not to invoke the `do` commands upon loading a game, or risk applying
side-effects multiple times.

In other words, doing it this way reduces side-effects, state, and corner cases.
So the architecture will be cleaner, and simpler even if makes for an
unfortunate disconnect between the mental model of game writers, and the mental
model of game developers.

### Model of Story Pages ###

#### Body ####

Given a gamestate, returns a list of resolved Strings to display to the player.
Each entry in the list represents a paragraph of text. Note that there should
be _no_ line breaks anywhere in the text. The UI willadd appropriate line
breaks when displaying the game content.

#### Choices ####

Returns a table that maps functions from gamestate to text to the next page.

#### Automatics ####

Standard automatics: Given a state, evaluates the predicates and returns the
first page whose predicate is satisfied by the state, or `nil` if there is
no such predicate.

## Combat Story Pages ##

Sometimes we will need to transition from telling a story to executing combat.

### Defining combat story pages ###

To transition from story to combat, we define an page with the following
format:

```yaml
- name: <EVENT_NAME>
  fight:
    isBoss: true/false
    enemies:
        armslength:
            -enemy_name1
            -enemy_name2
            -enemy_name1
        distant:
            -enemy_name4
            -enemy_name5
            -enemy_name6
    victory: page_name1
    defeat: page_name2
```

The `isBoss` field tells whether the battle is a boss battle or not. This
changes the music that is played.

The `enemies` field is where we define the enemies in the battle. They are
specified under two entries: `armslength` and `distant`, which specifies
the distance at which the enemies begin combat. Each field has a list of
the names of enemies to appear in combat.

Note that a name may appear multiple times. A separate table of that
character will be created for each entry.

### Internal model ###

#### Body ####

Returns a table of allies and enemies keyed on their current positions, and
their names. Note that the names are tweaked so that they are guaranteed to
be unique, and so that the gender of each enemy is made clear.

#### Choices ####

An empty table.

#### Automatics ####

An empty table

#### Transformer ####

Starts the combat music. Also creates the new copies of enemies for
combat (if those enemies are generic?).

## Combat Pages ##

There are actually two types of combat pages: `atomic` and `composite`.
An `atomic` combat page represents a single discrete action that an be
performed in combat, such as moving or attacking an opponent.

A `composite` combat page is made up of multiple combat pages. For example,
a power that hits multiple targets resolves into multiple attack pages, one
for each target.

There is a very small number of `atomic` combat pages, and these pages are
baked into the game engine. Meanwhile, there can be an unbounded number of
composite pages, and these may be defined by the game developer.

Combat itself is resolved in two passes:

1. Translate all of the composite pages into their appropriate atomic
pages. This may involve replacing a composite page with 1 or more atomic
pages, or iterating through the current chain of pages and adding/removing
pages.

2. Apply the atomic pages to execute the combat round.

Each page is the head of a recursive chain of pages. One iterates through the
chain via the `automatics` function (this allows us to execute the combat the
same way we execute any other automated sequence of pages).

### Atomic Combat Pages ###

There are four atomic combat pages:

1. `Damage` - Does (possibly negative) damage to a character
2. `Inflict` - Inflicts a status for a (possibly negative) duration on a
    character.
3. `Move` - Move a character to a different combat position.
4. `Message` - Generates a message to display to the player, but otherwise does
    nothing.

All the atomic pages have a function `message` function that takes the player
of the page, and the target of the page, and returns a String describing
the action. This function can be used to build a new page `B` from an
original page `A` that is identical except for the message.

Each page also has a special function with the same name as the page type
(i.e. `damage`, `move`, `inflict`). These functions take as arguments the
initiator of the action, and the target of the action and returns the
changes they will apply, as well as whether the changes are being applied to
an ally of the initiator. For example, `damage` returns the damage that will
be done to the target, `move` returns the position the target will move to, and
`inflict` returns the status that will be inflicted, and its duration.

#### Damage ####

Models doing damage to a character (healing can be accomplished by doing
negative damage).

##### Body #####

Given a combat state, returns a singleton list of strings describing who
is attacking who and for how much damage.

##### Choices #####

None

##### Automatics #####

Transitions to the next atomic combat page.

##### Transformer #####

Applies the damage to the recipient.

##### Damage #####

The `Damage` page has an additional entry `damage`, a function that
takes the user, and the target, and returns the damage done.

The `damage` and `message` functions can be used to derive new damage
pages from old pages.

#### Move ####

Models moving a character from one position to another.

##### Body #####

Same idea as [Damage](#damage).

##### Choices #####

None

##### Automatics #####

Next page in the sequence.

##### Transformer #####

Moves the character.

##### Move #####

Given a character, returns the position they are moving to.

#### Inflict ####

Inflicts a status on a character for a calculated duration. Curing a status
can be done by inflicting the status for a negative duration.

##### Body #####

Same idea as the other atomic actions.

##### Choices #####

None.

##### Automatics #####

Next page in the sequence.

##### Transformer #####

Inflicts this page's status on the target character.

##### Inflict #####

Given an initiator, and a target returns the status that will be inflicted, and
the status' duration.

#### Message ####

This action does nothing but prints a message. This can be useful if
a global message should be provided (or if a different page needs to be
transformed into a no-op). However, in order to describe the pages of a single
page `A`, `A.message` function should be used, so that if the page gets
modified by a complex page, the message function can be modified accordingly.

##### Body #####

Given the gamestate, returns a singleton list containing the message generated
by this page.

##### Choices #####

None.

##### Auotomatics #####

Returns the next page in the sequence.

##### Transformer #####

No-op.

### Complex Combat Pages ###

Complex combat pages are pages that change the current sequence of combat
pages. In most cases, a complex page will resolve into 1 or more atomic
pages that are appended to the page chain. However, complex pages can
also analyze the current chain and modify it.

For example, a `grapple` command may resolve into four atomic pages: one to
inflict a `grapple` status on the grappler, one to inflict a `grapple` status
on the grapplee (the `grapple` status locks both characters into the grapple
as long as it remains), one that moves the grappler into the grapple range,
and one that moves the grapplee into grapple range.

However, if a character defends another character, then the `defend` action
will go through and modify the other atomic actions so that those actions
that target the defended will instead target the defender.

To support this kind of analysis, `gamestate.combat` contains a table
`pages` that maps atomic pages to their user and target. It also contains
`start`, which maps to the first page in the chain, and `end` which
maps to the last page in the chain.

The only operation of interest in complex combat pages is the `transformer`,
which modifies the page chain.

#### Body ####

None.

#### Choices ####

None.

#### Automatics ####

The next complex page to resolve.

#### Transformer ####

Given a `gamestate`, updates `gamestate.combat.pages`,
`gamestate.combat.start` and `gamestate.combat.end` with the
atomic pages that this page resolves into.

#### Append ####

Most complex pages will be append pages, meaning that they resolve into one
or more atomic pages, which are appended to `gamestate.combat.end`.

To support this usecase, complex pages include a special function `append`.
`append` is invoked by the default implementation of `transformer`. It takes
the `gamestate`, and returns a table with the keys `start`, `pages` and
`end`. `start` is the head of the atomic page chain that this page
resolves into. `end` is the last atomic page that this page resolves into,
and `pages` is a mapping from atomic page to its user and target.
