--[[
-- Copyright 2016 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
-- A module for building and working with the various flavors of state used
-- in the game.
--]]

local M = {}

local customdata = require 'data/custom'
local scarlettables = require 'scarlettables'
local dynamics = require 'story/templatefunctions'
local predicates = require 'story/predicates'
local operations = require 'story/operations'
local storyload = require 'story/load'
local F = require "f-strings/F"

--[[
-- Returns the names of all the save files (*.sav) that currently exist.
--]]
local function existingsaves(love)
    local saveFiles = {}
    local numsaves = 1
    local savedirectory = love.filesystem.getSaveDirectory()
    for _, file in ipairs(love.filesystem.getDirectoryItems(savedirectory)) do
        if string.find("%.sav$", file) then
            saveFiles[numsaves] = file
            numsaves = numsaves + 1
        end
    end
    return saveFiles
end

function M.buildconfiguration()
    local configuration = {
        characters = customdata.characters,
        items = customdata.items,
        powers = customdata.powers,
        statuses = customdata.statuses,
        options = customdata.options,
        dynamics = dynamics,
        predicates = predicates,
        operations = operations
    }
    if customdata.stories then
        configuration.stories = customdata.stories 
    elseif GENERATE_HTML then
        configuration.stories = storyload.load(configuration)
    else
        configuration.stories = storyload.load(
            configuration, 
            "episode1_day1.story"
        )
    end
    if not configuration.options._keys then
        local keys = {}
        for name in pairs(configuration.options) do
            table.insert(keys, name)
        end
        configuration.options._keys = keys
    end
    return configuration
end

--[[
-- Loads the story file that contains the specified event into
-- `configuration.stories`.
--
-- configuration: A table containing a key `stories` that maps event 
-- names to events. `configuration.stories` will be populated with the
-- events in the loaded story file
-- name: The name of the event that needs to be loaded
--
--]]
local function loadevents(configuration, name)
    -- ep2d2_something or ep12d24_something_else
    local prefix_pattern = "ep(%d+)d(%d+)_.*"
    local _, _, episode, day = string.find(name, prefix_pattern)
    if not (episode and day) and string.find(name, "break_in_victory") then
        episode = 1
        day = 2
    end
    assert(
        episode and day, 
        F"Invalid format for event name: '{name}'."
        .. " Expects a pattern of the form ep<number>d<number>_<whatever>"
    )
    local filename = F"episode{episode}_day{day}.story"
    local events = storyload.load(configuration, filename)
    for name, event in pairs(events) do
        configuration.stories[name] = event
    end
end

--[[
-- Gets the story page with the specified name, loading the
-- file containing the specified story page if necessary.
--
-- configuration: The configuration that should contain the specified story event
-- name: The name of the event to fetch
--
-- return The story page to play
--]]
function M.getstory(configuration, name)
    if not configuration.stories[name] then
        loadevents(configuration, name)
    end
    if not configuration.stories[name] and string.find(name, "break_in_victory") then
        name = "ep1d2_" .. name
    end
    return assert(configuration.stories[name], F"Unknown event '{name}'.")
end

function M.buildgamestate(configuration)
    local gamestate = {
        configuration=configuration,
        keywords={},
        options={},
        characters=scarlettables.copy(configuration.characters),
        lossNumber=0,
        gametext={},
        visited={}
    }
    for _, optionname in ipairs(configuration.options._keys) do
        local option = assert(
            configuration.options[optionname],
            "Unknown option: " .. optionname
        )
        local startindex = option.startindex or 1
        local defaultoption = option[startindex]
        defaultoption.apply(gamestate)
        gamestate.options[optionname] = defaultoption.name
    end
    return gamestate
end

--[[
-- For the console version, we don't have access to Love's filesystem
-- utilities, and Lua has very limited filesystem utilities. So we provide
-- a fixed number of nine save slots.
--]]
function saveslots()
    local slots = {}
    -- We should do something smarter here where we display the name of 
    -- the character, and the current episode they're on or something.
    for i=1,9 do
        table.insert(slots, "slot " .. i)
    end
end

--[[
-- Builds the initial UI state. It is here that we load game-specific data
-- from the data directives.
--]]
function M.builduistate(gamestate)
    local love = LOVE and require 'love'
    local suit = LOVE and require 'SUIT'
    local uistate = {
        gui=suit,
        core={gamestate=gamestate},
        save={
            oldsave=1,
            newsave=1,
            textinput={text=""},
            allsaves=LOVE and existingsaves(love) or saveslots()
        }
    }
    return uistate
end

return M
