--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]
--
--[[
-- Defines a page that contains a list of buttons the player can click on
-- to move on to the next page.
--]]
--
local M = {}

local fonts = require 'ui/fonts'
local pages = require 'pages'
local storyutils = require 'story/storyutils'

-- [[
-- Returns the number of characters that can fit on one line of text.
--
-- font: The font for which we want to know the number of characters that will
--  fit
-- screenwidth: The width of the screen in pixels
--
-- returns The number of characters that will fit on a line with this font
-- ]]
local function numcharsperline(font, screenwidth)
    local charwidth = fonts.singleCharacterWidth(fonts.body(bodyfontsize))
    local screenwidth = love.graphics.getWidth()
    local screenlinelength = math.floor((screenwidth - charwidth) / charwidth)
    local numchars = (screenwidth - charwidth) / charwidth
    if numchars > storyutils.MAX_LINE_LENGTH then
        numchars = storyutils.MAX_LINE_LENGTH
    end
    return numchars
end

local function drawgui(uistate, uichoices, choicestext)
    local gui = uistate.gui
    local titlefont= fonts.title(uistate.core.gamestate.options.titlefont)
    local titletext = "Choices:"
    gui.Label(
        titletext,
        {font=titlefont},
        gui.layout:row(titlefont:getWidth(titletext) + 5, titlefont:getHeight())
    )
    local numcharsline = numcharsperline(bodyfont, screenwidth)
    local wrappedchoices = storyutils.wraptext(choicestext, numcharsline)
    local buttonfont = fonts.button(uistate.core.gamestate.options.buttonfont)
    for i, choicetext in ipairs(wrappedchoices) do
        local uikey = tostring(i)
        local choice = uichoices[uikey]
        local concatenatedtext = table.concat(choicetext, "\n")
        local buttontext = uikey .. ": " .. concatenatedtext
        local choicebuttonwidth = buttonfont:getWidth(buttontext)
        gui.Button(
            buttontext,
            {font=buttonfont, id=uikey, align="left", valign="top"},
            gui.layout:row(choicebuttonwidth + 5, 2 * #choicetext * buttonfont:getHeight())
        )
    end
    local backText = "<==Return to Story"
    gui.Button(
        backText,
        {font=titlefont, id="backspace"},
        gui.layout:row(titlefont:getWidth(backText) + 5, titlefont:getHeight())
    )
end

local function choices(uichoices, gamestate)
    local copy = {_keys={}}
    for _, key in ipairs(uichoices._keys) do
        copy[key] = uichoices[key]
        table.insert(copy._keys, key)
    end
    local storyui = require 'story/storyui'
    copy.backspace = storyui.page(gamestate.storypagename)
    table.insert(copy._keys, "backspace")
    return copy
end

function M.buttons(uichoices, choicestext)
    local choicebuttons = pages.page(
        nil,
        nil,
        function(uistate)
            return choices(uichoices, uistate.core.gamestate)
        end,
        nil
    )
    function choicebuttons.drawgui(uistate)
        local gamestate = uistate.core.gamestate
        drawgui(uistate, choices(uichoices, gamestate), choicestext)
    end
    return choicebuttons
end


return M
