--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]
-- [[
-- This module contains the basic functions that can be used as building blocks for
-- dynamic functions used in story text.
-- ]]

local M = {}

--[[
-- This function displays text depending on whether the specified character has
-- the specified value for the specified attribute. For example,
-- `attribute;greg; gender; male; he; she` will return `he` if the `greg`'s
-- gender is `male`, and will return `she` otherwise.

-- If only the character and attribute name are specified, then this function
-- will return the value the specified character has for the specified attribute.
-- For example, suppose `greg` is `male`. Then, `attribute(greg; gender)` will
-- return `male`.
--
-- character - The character whose attribute needs to be checked
-- attribute - The name of the attribute to check
-- value - The value to compare the character's attribute against
-- textifequal - The text to print of the character's attribute is equal to
-- `value`, defaults to the empty string.
-- textotherwise - The text to print otherwise, defaults to the empty string
--]]
function M.attribute(character, attribute, value, textifequal, textotherwise)
    if not value then
        return character.attributes[attribute]
    end
    return string.lower(value) == string.lower(character.attributes[attribute])
        and (textifequal or "")
        or (textotherwise or "")
end

--[[
-- This function will print text depending on whether or not the player has
-- acquired the specified [keyword](#keywords). For example, the function
-- `keyword;the_thing; I've got the thing!; I don't have the thing...`
-- will return `I've got the thing!'` if the player has the keyword
-- `the_thing` and `I don't have the thing` if the player does not have
-- the keyword `the_thing`.
--
-- possessedkeywords - The set of keywords possessed by the player
-- keyword - The keyword to look for
-- textifhaskeyword - The text to return if the player has the specified keyword, defaults to
--  the empty string
-- textotherwise - The text to return if the player does not have the specified keyword,
--  defaults to the empty string
--
-- return Either textifhaskeyword or textotherwise depending on whether the
-- player has the specified keyword or not
-- ]]
function M.keyword(
    possessedkeywords,
    keyword,
    textifhaskeyword,
    textotherwise
)
    textifhaskeyword = textifhaskeyword or ""
    textotherwise = textotherwise or ""
    return possessedkeywords[keyword] and textifhaskeyword or textotherwise
end

--[[
-- This function will print text depending on the relative value of a statistic.
-- For example, `statistic;greg; hunger; gt; 5; I am hungry!; I am not...` will
-- return `I am hungry!` if `greg`'s `hunger` is greater than `5`, and
-- `I am not...` otherwise.

-- If only the `character` and `statisticname` are specified, then this function
-- returns the current value of the specified statistic. So, suppose `greg` has
-- a `hunger` of `5`. Then, `statistic(greg; hunger)` returns `5`.
--
-- character: The character whose statistic should be checked
-- statisticname: The name of the statistic to check
-- comparison: The comparison to be performed, one of `gt` (greater than), `lt` (less than),
--  `e` (equal), `gte` (greater than or equal), or `lte` (less than or equal)
--  value: The value to compare the character's statistic against
--  textiftrue: The text to return if the comparison is true, defaults to the empty string
--  textiffalse: The text to return if the comparison is false, defaults to the empty string
--]]
function M.statistic(
    character,
    statisticname,
    comparison,
    value,
    textiftrue,
    textiffalse
)
    textiftrue = textiftrue or ""
    textiffalse = textiffalse or ""
    local statisticvalue = character.statisticvalue[statisticname]

    if not value then
        return tostring(statisticvalue)
    end

    local checkpasses = false
    if comparison == "gt" then
        checkpasses = statisticvalue > value
    elseif comparison == "lt" then
        checkpasses = statisticvalue < value
    elseif comparison == "gte" then
        checkpasses = statisticvalue >= value
    elseif comparison == "lte" then
        checkpasses = statisticvalue <= value
    end

    return checkpasses and textiftrue or textiffalse
end

return M
