--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
-- Defines the operations available for use in the `do` block in story events.
--
-- Note that the functions defined in this module are *not* the operations.
-- Instead, they're functions that *return* the operation on gamestate that
-- needs to be executed.
--]]

local log = require 'log'
local M = {}
local operations = require 'story/atomicoperations'
local powers = require 'data/powers'

local function getcharacter(gamestate, name)
    return assert(
        gamestate.characters[name],
        "Unknown character: " .. name
    )
end

--[[
-- Returns an operation on gamestate that changes the specified attribute on the specified
-- character to the
-- specified value. For example `attribute(greg, gender, female)` will make `greg`
-- a woman.
--
-- attributename - The name of the attribute to be changed
-- value - The value to change the attribute to
-- name - The name of the character whose attribute should be changed, defaults to "pc"
--]]
function M.attribute(attributename, newvalue, name)
    name = name or "pc"
    log.trace("Attribute arguments: ", name, attributename, newvalue)
    return function(gamestate)
        operations.attribute(gamestate.characters[name], attributename, newvalue)
    end
end

--[[
-- Returns an operation on gamestate that sets the specified statistic on the specified
-- character to the
-- specified value. For example `statistic(greg, hunger, 10)` sets `greg`'s
-- hunger statistic to 10.
--
-- statisticname: The name of the statistic to be modified
-- newvalue: The newvalue for the statistic
-- name: The name of the character whose statistic needs to be changed
--]]
function M.statistic(statisticname, newvalue, name)
    name = name or "pc"
    log.trace("Statistic arguments: ", name, statisticname, newvalue)
    local numericvalue = assert(
        tonumber(newvalue),
            "Can't set " .. name .. "'s " .. statisticname .. " to non-numeric value '"
            .. newvalue .. "'"
    )
    return function(gamestate)
        operations.statistic(gamestate.characters[name], statisticname, numericvalue)
    end
end

--[[
-- Returns an operation on gamestate that adds the specified keyword to the player's set of
-- acquired keywords.
--
-- newkeyword- The keyword to add to the player's set of keywords
--]]
function M.keyword(newkeyword)
    log.trace("Adding keyword: ", newkeyword)
    return function(gamestate)
        operations.keyword(gamestate.keywords, newkeyword)
    end
end

--[[
-- Returns an operation on gamestate that increments the specified character's specified
-- statistic by the specified amount. If
-- no amount is specified, then the statistic is incremented by 1.
--
-- statisticname: The name of the statistic to amount
-- amount: The amount to amount the statisic by, defaults to one
-- name: The name of the character whose statistic should be amounted, defaults to "pc"
--]]
function M.incrementStatistic(statisticname, amount, name)
    amount = amount or 1
    name = name or "pc"
    log.trace("Incrementing statistic: ", name, statisticname, amount)
    local errorMessage = "Can't set "
        .. name .. "'s "
        .. statisticname
        .. " to "
        .. "non numeric value '"
        .. amount .. "'"
    local numericamount = assert(tonumber(amount), errorMessage)
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        local newvalue
        if statisticname == "energy" then
            local newenergy = character.statistics.energy + numericamount
            newvalue = math.min(character.statistics.maxenergy, newenergy)
        else
            newvalue = character.statistics[statisticname] + numericamount
        end
        operations.statistic(character, statisticname, newvalue)
    end
end

--[[
-- Returns an operation on gamestate that decrements the specified character's specified
-- statistic by the specified amount. If
-- no amount is specified, then the statistic is decremented by 1.
--
-- statisticname: The name of the statistic to decrement
-- amount: The amount to decrement the statisic by, defaults to one
-- name: The name of the character whose statistic should be decremented
--]]
function M.decrementStatistic(statisticname, amount, name)
    name = name or "pc"
    return M.incrementStatistic(
        statisticname,
        amount and "-" .. amount or -1,
        name
    )
end

-- [[
--  Returns an operation on gamestate that teaches the specified power to
--  the specified character.
--
--  powername: The name of the power to learn
--  name: The name of the character learning the power, defaults to pc
-- ]]
function M.learn(powername, name)
    name = name or "pc"
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        local power = assert(
            powers[powername],
            "Unknown power: " .. "'" .. powername .. "'"
        )
        if not character.powers[powername] then
            character.powers[powername] = power
            table.insert(character.powers._keys, powername)
        end
    end
end

function M.equip(itemname, name)
    name = name or "pc"
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        local clothing = assert(
            gamestate.configuration.items[itemname],
            "Unknown item: " .. itemname
        )
        for _, slot in pairs(clothing.slots) do
            local oldclothing = character.clothing[slot]
            if oldclothing then
                for _, oldslot in pairs(oldclothing.slots) do
                    character.clothing[oldslot] = nil
                end
            end
            character.clothing[slot] = clothing
        end
    end
end

function M.equipWorkClothes(name)
    name = name or "pc"
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        if character.gender == "female" then
            M.equip("waitress skirt", name)(gamestate)
            M.equip("waitress shirt", name)(gamestate)
        else
            M.equip("waiter pants", name)(gamestate)
            M.equip("waiter shirt", name)(gamestate)
        end
    end
end

function M.equipGenderClothing(name)
    name = name or "pc"
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        M.equip("jeans", name)(gamestate)
        M.equip("t-shirt", name)(gamestate)
        if character.gender == "female" then
            M.equip("lacy panties", name)(gamestate)
        else
            M.equip("boxer briefs", name)(gamestate)
        end
    end
end

function M.equipSexyWorkClothes(name)
    name = name or "pc"
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        if character.attributes.gender == "female" then
            M.equip("short waitress skirt", name)(gamestate)
            M.equip("waitress vest", name)(gamestate)
        else
            M.equip("tight waiter pants", name)(gamestate)
            M.equip("waiter vest", name)(gamestate)
        end
    end
end

local function alreadyWearingCostume(character)
    local outfit = character.clothing
    local aslot = character.clothing._keys[1]
    return outfit[aslot] == "scarlet thong" or outfit[aslot] == "scarletsuit"
end

function M.equipCostume(name)
    name = name or "pc"
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        if alreadyWearingCostume(character) then
            return
        end
        local originaloutfit = {}
        for _, slot in pairs(character.clothing._keys) do
            originaloutfit[slot] = character.clothing[slot]
        end
        local costumename = gamestate.keywords.sexyOutfit
            and "scarlet thong"
            or "scarletsuit"
        M.equip(costumename, name)(gamestate)
        gamestate.originaloutfit = originaloutfit
    end
end

function M.hairstyle(style, name)
  name = name or "pc"
  return function(gamestate)
    local character = getcharacter(gamestate, name)
    character.attributes.hairstyle = style
  end
end

function M.removeCostume(name)
    name = name or "pc"
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        local originaloutfit = gamestate.originaloutfit
        if not originaloutfit then
            return
        end
        for slot, clothing in pairs(originaloutfit) do
            character.clothing[slot] = clothing
        end
        gamestate.originaloutfit = nil
    end
end

function M.sore(marks, name)
    return M.bummarks(marks, name)
end

function M.bummarks(marks, name)
    name = name or "pc"
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        local bummarks = character.bummarks
        if not character.bummarks then
          character.bummarks = {}
        end
        table.insert(character.bummarks, marks)
    end
end

function M.bum(marks, name)
    return M.bummarks(marks, name)
end

function M.clearBummarks(name)
    name = name or "pc"
    return function(gamestate)
        getcharacter(gamestate, name).bummarks = nil
    end
end

-- Deprecated. Use M.healBum instead.
function M.healBummark(name)
    return M.clearBummarks(name)
end

-- Deprecated. Use M.healBum instead.
function M.healBummarks(name)
    return M.clearBummarks(name)
end

function M.partialHealBum(name, numheals)
    name = name or "pc"
    numheals = numheals or 1
    return function(gamestate)
    local bummarks = getcharacter(gamestate, name).bummarks
    if bummarks then
        for i=1,numheals do
            table.remove(bummarks)
        end
    end
    end
end

function M.healBum(name)
    return M.clearBummarks(name)
end

function M.setBattleName(battlename, name)
    name = name or "pc"
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        character.combatname = battlename
        character.concisename = string.sub(character.combatname, 1, 2)
    end
end

function M.heal(name)
    name = name or "pc"
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        character.statistics.energy = character.statistics.maxenergy
    end
end

return M
