--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]
--
--[[
-- This module defines the predicates that can be used as part of the automatics block for
-- a story event.
--]]

local M = {}

local predicates = require 'story/atomicpredicates'

local function getcharacter(gamestate, name)
    return assert(gamestate.characters[name], "Unknown character: " .. name)
end

--[[
-- Given an attribute name, a value and (optionally) a character's name, returns a
-- predicate on gamestate that returns true if the specified character's specified attribute
-- has the specified value.
--
-- attributename: The name of the attribute to check
-- value: The value to check for
-- character: The name of the character whose attribute should be checked, defaults to 'pc'
--]]
function M.attribute(attributename, value, character)
    character = character or 'pc'
    return function(gamestate)
        return predicates.attribute(
            gamestate.characters[character],
            attributename,
            value
        )
    end
end

function M.love()
    return function() return LOVE end
end

function M.keyword(keyword)
    return function(gamestate)
        return gamestate.keywords[keyword]
    end
end

function M.isfemale(name)
    name = name or "pc"
    return function(gamestate)
        return getcharacter(gamestate, name).attributes.gender == "female"
    end
end

function M.female(name)
    return M.isfemale(name)
end

function M.femaleclothing(name)
    return M.female(name)
end


function M.ismale(name)
    name = name or "pc"
    return function(gamestate)
        return getcharacter(gamestate, name).attributes.gender == "male"
    end
end

function M.bummarks(name)
    name = name or "pc"
    return function(gamestate)
        local bummarks = getcharacter(gamestate, name).bummarks
        return bummarks and next(bummarks)
    end
end

function M.statistic(statisticname, comparison, value, character)
    character = character or "pc"
    local numericvalue = assert(
        tonumber(value),
        "Cannot compare statistic "
            .. statisticname
            .. " to non-numeric value "
            .. value
    )
    return function(gamestate)
        return predicates.statistic(
            gamestate.characters[character],
            statisticname,
            comparison,
            numericvalue
        )
    end
end

M["true"] = function()
    return function() return true end
end

M.otherwise = M["true"]

function M.equal(statistic, value, character)
    return M.statistic(statistic, "e", value, character)
end

function M.greaterThan(statistic, value, character)
    return M.statistic(statistic, "gt", value, character)
end

function M.greaterthan(statistic, value, character)
    return M.greaterThan(statistic, value, character)
end

function M.gt(statistic, value, character)
    return M.greaterThan(statistic, value, character)
end


function M.lessThan(statistic, value, character)
    return M.statistic(statistic, "lt", value, character)
end

function M.lessthan(statistic, value, character)
    return M.lessThan(statistic, value, character)
end

function M.lt(statistic, value, character)
    return M.lessThan(statistic, value, character)
end

function M.playernotspankable()
    return function(gamestate)
        return gamestate.options.spankableplayer == "no"
    end
end

function M.playerspankable()
    return function(gamestate)
        return gamestate.options.spankableplayer == "yes"
    end
end

function M.wearsclothing(name, slot, ...)
    name = name or "pc"
    local types = {...}
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        for _, type in ipairs(types) do
            if character.clothing[slot].type == type then
                return true
            end
        end
        return false
    end
end

function M.wearsPantsOrShorts(name)
    return M.wearsclothing(name, "lower", "pants", "shorts")
end

function M.wearsSkirtOrDress(name)
    return M.wearsclothing(name, "lower", "skirt", "dress")
end

function M.underwearIsBaring(name)
    name = name or "pc"
    return function(gamestate)
        local underwear = getcharacter(gamestate, name).clothing.underwear
        return underwear and underwear.risque >= 5
    end
end

function M.wearsUnderwear(name)
    name = name or "pc"
    return function(gamestate)
        return getcharacter(gamestate, name).clothing.underwear ~= nil
    end
end

local function visible(slot, clothing)
    return slot ~= "underwear" or clothing["lower"]
end

function M.scantilyClad(name)
    name = name or "pc"
    return function(gamestate)
        local risque = 0
        local seenclothing = {}
        local character = getcharacter(gamestate, name)
        for _, slot in pairs(character.clothing._keys) do
            if visible(slot, character.clothing) then
                local outfit = character.clothing[slot]
                if not seenclothing[outfit] then
                    risque = risque + outfit.risque
                    seenclothing[outfit] = 1
                end
            end
        end
        return risque > 3
    end
end

local SPANKER = 1
local SPANKEE = -1
local SWITCH = 0
local function spankerspankeeswitch(character)
    local spanker = character.statistics.spanker or 1
    local spankee = character.statistics.spankee or 1
    if spanker == spankee then
        return SWITCH
    end
    return spanker >= 2 * spankee
        and SPANKER
        or spankee >= 2 * spanker
        and SPANKEE
        or SWITCH
end

function M.role(name, role)
    name = name or "pc"
    return function(gamestate)
        return role == spankerspankeeswitch(getcharacter(gamestate, name))
    end
end

function M.spanker(name)
    return M.role(name, SPANKER)
end

function M.spankee(name)
    return M.role(name, SPANKEE)
end

function M.switch(name)
    return M.role(name, SWITCH)
end

local COMPETENT = 1
local INCOMPETENT = -1
local BALANCED  = 0

local function competentincompetentbalanced(character)
    local victory = character.statistics.victory
    local defeat = character.statistics.defeat
    -- Handles the "both are zero" edge case.
    if victory == defeat then
        return BALANCED
    end
    return victory >= 2 * defeat
        and COMPETENT
        or defeat >= 2 * victory
        and INCOMPETENT
        or BALANCED
end

function M.competence(ability, name)
    name = name or "pc"
    return function(gamestate)
        local character = getcharacter(gamestate, name)
        return ability == competentincompetentbalanced(character)
    end
end

function M.competent(name)
    return M.competence(COMPETENT, name)
end

function M.inccompetent(name)
    return M.competence(INCOMPETENT, name)
end

function M.balanced(name)
    return M.competence(BALANCED, name)
end

function M.hairlength(length, name)
    return M.attribute("hairlength", length, name or "pc")
end

return M
