--[[
-- Copyright 2016 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- uiutilitiesERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
-- Contains a collection of utility functions to simplify building UI pages.
--]]
--

local uiutilities = {}

local buttons = LOVE and require 'ui/buttons'
local pages = require 'pages'
local scarlettables = require 'scarlettables'

local utilities = require 'utilities'

local suit = LOVE and require 'SUIT'

uiutilities.up = utilities.up
uiutilities.down = utilities.down
uiutilities.right = utilities.right
uiutilities.left = utilities.left
uiutilities.backspace = utilities.backspace

--[[
--Wraps the suit.layout:col to simplify passing it around as a
--zero-argument function. Useful when building functions to build similar
--widgits that need to be laid out in slightly different relative positions.
--]]
function uiutilities.col()
    return suit.layout:col()
end

--[[
-- Wraps the suit.layout:row to simplify passing it around as zero-argument
-- function. Useful when building functions to build similar widgits that need
-- to be laid out in slightly different relative positions
--]]
function uiutilities.row()
    return suit.layout:row()
end

--[[
-- Builds a back button. Note that in order for a back button to function,
-- the current page must be made backable
--]]
function uiutilities.buildBackButton(uistate)
    return buttons.buildButton(
        uistate,
        "(<==) Back",
        "backspace",
        uiutilities.row
    )
end

--[[
-- The back page allows the user to undo the previous action.
--
-- The back page has no body, and its transformer restores the current state
-- to the state of the page before the undoable page.
-- Then it automatically transitions to the newly
-- restored state's current page (which is the page that executed just
-- before the page that supports undo operations).
--
-- Note that this page assumes that page.core.previous is not nil.
--
-- transformer: An optional transformer to apply after restoring the old
-- core.
--]]
function uiutilities.backpage(transformer)
    return pages.page(
            function() end,
            nil,
            nil,
            function(uistate)
                uistate.core = uistate.core.previous
                if transformer then transformer(uistate) end
            end
    )
end

--[[
-- This operation makes a state backable, by making a deep copy of the
-- core of the current state, and storing that copy as the core.previous.
-- It also sets the core.previous.state to previousPage in order to give
-- us a copy of the state from _just before_ the page transition occurs.
--
-- Note: This is intended to be passed in directly as the transformer for
-- a backable UI page. The main.lua actually passes in more than just the
-- uistate (in particular it also passes in the previousPage).
--]]
function uiutilities.makePageBackable(uistate, previousPage)
    uistate.core.previous = scarlettables.copy(uistate.core)
    uistate.core.previous.page = previousPage
end

-- [[
-- Wipes out all saved previous states, effectively resetting the undo.
--
-- uistate: The state whose previous copies need to be wiped out
-- ]]
function uiutilities.clearundo(uistate)
    uistate.core.previous = nil
end

return uiutilities
